# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Ethereum module wrapping the public and private key cryptography and ledger api."""

import decimal
import json
import logging
import math
import threading
import warnings
from copy import deepcopy
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union, cast
from uuid import uuid4

import ipfshttpclient  # noqa: F401 # pylint: disable=unused-import
import web3._utils.request
from eth_account import Account
from eth_account._utils.signing import to_standard_signature_bytes
from eth_account.datastructures import HexBytes, SignedTransaction
from eth_account.messages import _hash_eip191_message, encode_defunct
from eth_account.signers.local import LocalAccount
from eth_keys import keys
from eth_typing import HexStr
from eth_utils.currency import from_wei, to_wei  # pylint: disable=import-error
from requests import HTTPError
from web3 import HTTPProvider, Web3
from web3._utils.request import SimpleCache
from web3.datastructures import AttributeDict
from web3.exceptions import ContractLogicError, TransactionNotFound
from web3.gas_strategies.rpc import rpc_gas_price_strategy
from web3.middleware import geth_poa_middleware
from web3.types import TxData, TxParams, TxReceipt, Wei

from aea.common import Address, JSONLike
from aea.crypto.base import Crypto, FaucetApi, Helper, LedgerApi
from aea.crypto.helpers import DecryptError, KeyIsIncorrect, hex_to_bytes_for_key
from aea.exceptions import enforce
from aea.helpers import http_requests as requests
from aea.helpers.base import try_decorator
from aea.helpers.io import open_file


_default_logger = logging.getLogger("aea.ledger_apis.ethereum")

_ETHEREUM = "ethereum"
TESTNET_NAME = "ganache"
DEFAULT_ADDRESS = "http://127.0.0.1:8545"
DEFAULT_CHAIN_ID = 1337
DEFAULT_CURRENCY_DENOM = "wei"
ETH_GASSTATION_URL = "https://ethgasstation.info/api/ethgasAPI.json"
_ABI = "abi"
_BYTECODE = "bytecode"
EIP1559 = "eip1559"
EIP1559_POLYGON = "eip1559_polygon"
GAS_STATION = "gas_station"
AVAILABLE_STRATEGIES = (EIP1559, GAS_STATION, EIP1559_POLYGON)
SPEED_FAST = "fast"  # safeLow, standard, fast
POLYGON_GAS_ENDPOINT = "https://gasstation-mainnet.matic.network/v2"
MAX_GAS_FAST = 1500

# How many blocks to consider for priority fee estimation
FEE_HISTORY_BLOCKS = 10

# Which percentile of effective priority fees to include
FEE_HISTORY_PERCENTILE = 5

# Which base fee to trigger priority fee estimation at (GWEI)
PRIORITY_FEE_ESTIMATION_TRIGGER = 100

# Returned if above trigger is not met (GWEI)
DEFAULT_PRIORITY_FEE = 3

# In case something goes wrong fall back to this estimate
FALLBACK_ESTIMATE = {
    "maxFeePerGas": to_wei(20, "gwei"),
    "maxPriorityFeePerGas": to_wei(DEFAULT_PRIORITY_FEE, "gwei"),
}

PRIORITY_FEE_INCREASE_BOUNDARY = 200  # percentage

DEFAULT_EIP1559_STRATEGY = {
    "max_gas_fast": MAX_GAS_FAST,
    "fee_history_blocks": FEE_HISTORY_BLOCKS,
    "fee_history_percentile": FEE_HISTORY_PERCENTILE,
    "priority_fee_estimation_trigger": PRIORITY_FEE_ESTIMATION_TRIGGER,
    "default_priority_fee": DEFAULT_PRIORITY_FEE,
    "fallback_estimate": FALLBACK_ESTIMATE,
    "priority_fee_increase_boundary": PRIORITY_FEE_INCREASE_BOUNDARY,
}

DEFAULT_EIP1559_STRATEGY_POLYGON = {
    "gas_endpoint": POLYGON_GAS_ENDPOINT,
    "speed": SPEED_FAST,
    "fallback_estimate": FALLBACK_ESTIMATE,
}

DEFAULT_GAS_STATION_STRATEGY = {"gas_price_api_key": "", "gas_price_strategy": "fast"}

GAS_STATION_FALLBACK_ESTIMATE = 20  # gwei

DEFAULT_GAS_PRICE_STRATEGIES = {
    EIP1559: DEFAULT_EIP1559_STRATEGY,
    GAS_STATION: DEFAULT_GAS_STATION_STRATEGY,
    EIP1559_POLYGON: DEFAULT_EIP1559_STRATEGY_POLYGON,
}

# The tip increase is the minimum required of 10%.
TIP_INCREASE = 1.1


def wei_to_gwei(number: Type[int]) -> Union[int, decimal.Decimal]:
    """Covert WEI to GWEI"""
    return from_wei(cast(int, number), unit="gwei")


def round_to_whole_gwei(number: Type[int]) -> Wei:
    """Round WEI to equivalent GWEI"""
    gwei = wei_to_gwei(number)
    rounded = math.ceil(gwei)
    return cast(Wei, to_wei(rounded, "gwei"))


def get_base_fee_multiplier(base_fee_gwei: int) -> float:
    """Returns multiplier value."""

    if base_fee_gwei <= 40:  # pylint: disable=no-else-return
        return 2.0
    elif base_fee_gwei <= 100:  # pylint: disable=no-else-return
        return 1.6
    elif base_fee_gwei <= 200:  # pylint: disable=no-else-return
        return 1.4
    else:  # pylint: disable=no-else-return
        return 1.2


def estimate_priority_fee(
    web3_object: Web3,
    base_fee_gwei: int,
    block_number: int,
    priority_fee_estimation_trigger: int,
    default_priority_fee: int,
    fee_history_blocks: int,
    fee_history_percentile: int,
    priority_fee_increase_boundary: int,
) -> Optional[int]:
    """Estimate priority fee from base fee."""

    if base_fee_gwei < priority_fee_estimation_trigger:
        return default_priority_fee

    fee_history = web3_object.eth.fee_history(
        fee_history_blocks, block_number, [fee_history_percentile]  # type: ignore
    )

    # This is going to break if more percentiles are introduced in the future,
    # i.e., `fee_history_percentile` param becomes a `List[int]`.
    rewards = sorted([reward[0] for reward in fee_history["reward"] if reward[0] > 0])
    if len(rewards) == 0:
        return None

    # Calculate percentage increases from between ordered list of fees
    percentage_increases = [
        ((j - i) / i) * 100 for i, j in zip(rewards[:-1], rewards[1:])
    ]
    highest_increase = max(*percentage_increases)
    highest_increase_index = percentage_increases.index(highest_increase)

    values = rewards.copy()
    # If we have big increase in value, we could be considering "outliers" in our estimate
    # Skip the low elements and take a new median
    if (
        highest_increase > priority_fee_increase_boundary
        and highest_increase_index >= len(values) // 2
    ):
        values = values[highest_increase_index:]

    return values[len(values) // 2]


def get_gas_price_strategy_eip1559(
    max_gas_fast: int,
    fee_history_blocks: int,
    fee_history_percentile: int,
    priority_fee_estimation_trigger: int,
    default_priority_fee: int,
    fallback_estimate: Dict[str, Optional[int]],
    priority_fee_increase_boundary: int,
) -> Callable[[Web3, TxParams], Dict[str, Wei]]:
    """Get the gas price strategy."""

    def eip1559_price_strategy(
        web3: Web3,  # pylint: disable=redefined-outer-name
        transaction_params: TxParams,  # pylint: disable=unused-argument
    ) -> Dict[str, Wei]:
        """
        Get gas price using EIP1559.

        Visit `https://github.com/ethereum/EIPs/blob/master/EIPS/eip-1559.md`
        for more information.

        :param web3: web3 instance
        :param transaction_params: transaction parameters
        :return: dictionary containing gas price strategy
        """

        latest_block = web3.eth.get_block("latest")
        base_fee = latest_block.get("baseFeePerGas")
        block_number = latest_block.get("number")
        base_fee_gwei = wei_to_gwei(base_fee)

        estimated_priority_fee = estimate_priority_fee(
            web3,
            cast(int, base_fee_gwei),
            block_number,
            priority_fee_estimation_trigger=priority_fee_estimation_trigger,
            default_priority_fee=default_priority_fee,
            fee_history_blocks=fee_history_blocks,
            fee_history_percentile=fee_history_percentile,
            priority_fee_increase_boundary=priority_fee_increase_boundary,
        )

        if estimated_priority_fee is None:
            _default_logger.warning(
                "An error occurred while estimating priority fee, falling back"
            )
            return fallback_estimate

        max_priority_fee_per_gas = max(
            estimated_priority_fee, to_wei(default_priority_fee, "gwei")
        )
        multiplier = get_base_fee_multiplier(cast(int, base_fee_gwei))

        potential_max_fee = base_fee * multiplier
        max_fee_per_gas = (
            (potential_max_fee + max_priority_fee_per_gas)
            if max_priority_fee_per_gas > potential_max_fee
            else potential_max_fee
        )

        if (
            wei_to_gwei(max_fee_per_gas) >= max_gas_fast
            or wei_to_gwei(max_priority_fee_per_gas) >= max_gas_fast
        ):
            return fallback_estimate

        return {
            "maxFeePerGas": round_to_whole_gwei(max_fee_per_gas),
            "maxPriorityFeePerGas": round_to_whole_gwei(max_priority_fee_per_gas),
        }

    return eip1559_price_strategy


def get_gas_price_strategy_eip1559_polygon(
    gas_endpoint: str,
    fallback_estimate: Dict[str, Optional[int]],
    speed: Optional[str] = SPEED_FAST,
) -> Callable[[Any, Any], Dict[str, Wei]]:
    """Get the gas price strategy."""

    def eip1559_price_strategy(
        web3: Web3,  # pylint: disable=redefined-outer-name,unused-argument
        transaction_params: TxParams,  # pylint: disable=unused-argument
    ) -> Dict[str, Wei]:
        try:
            response = requests.get(gas_endpoint)
            if response.status_code == 200:
                data = response.json()[speed]
                return {
                    "maxFeePerGas": Wei(to_wei(data["maxFee"], "gwei")),
                    "maxPriorityFeePerGas": Wei(to_wei(data["maxPriorityFee"], "gwei")),
                }
            return fallback_estimate
        except requests.exceptions.RequestException:
            return fallback_estimate

    return eip1559_price_strategy


def rpc_gas_price_strategy_wrapper(
    web3: Web3, transaction_params: TxParams  # pylint: disable=redefined-outer-name
) -> Dict[str, Wei]:
    """RPC gas price strategy wrapper."""
    return {"gasPrice": rpc_gas_price_strategy(web3, transaction_params)}


def get_gas_price_strategy(
    gas_price_strategy: Optional[str] = None, gas_price_api_key: Optional[str] = None
) -> Callable[[Web3, TxParams], Dict[str, Wei]]:
    """Get the gas price strategy."""
    supported_gas_price_modes = ["safeLow", "average", "fast", "fastest"]

    if gas_price_strategy is None:
        _default_logger.debug(
            "Gas price strategy not provided. Falling back to `rpc_gas_price_strategy`."
        )
        return rpc_gas_price_strategy_wrapper

    if gas_price_strategy not in supported_gas_price_modes:
        _default_logger.debug(
            f"Gas price strategy `{gas_price_strategy}` not in list of supported modes: {supported_gas_price_modes}. Falling back to `rpc_gas_price_strategy`."
        )
        return rpc_gas_price_strategy_wrapper

    if gas_price_api_key is None:
        _default_logger.debug(
            "No ethgasstation api key provided. Falling back to `rpc_gas_price_strategy`."
        )
        return rpc_gas_price_strategy_wrapper

    def gas_station_gas_price_strategy(
        web3: Web3,  # pylint: disable=redefined-outer-name
        transaction_params: TxParams,  # pylint: disable=unused-argument
    ) -> Dict[str, Wei]:
        """
        Get gas price from Eth Gas Station api.

        Visit `https://docs.ethgasstation.info/gas-price` for documentation.
        :param web3: web3 instance
        :param transaction_params: transaction parameters
        :return: wei
        """
        _default_logger.info(  # pragma: nocover
            "`ethgasstation.info` has been deprecated and will be replaced with an alternative on the next release."
        )
        response = requests.get(f"{ETH_GASSTATION_URL}?api-key={gas_price_api_key}")
        if response.status_code != 200:  # pragma: nocover
            # TODO : Use some other gas station API # pylint: disable=fixme
            _default_logger.error(
                f"Gas station API response: {response.status_code}, {response.text}, using fallback gas price."
            )
            return {"gasPrice": web3.to_wei(GAS_STATION_FALLBACK_ESTIMATE, "gwei")}
        response_dict = response.json()
        _default_logger.debug("Gas station API response: {}".format(response_dict))
        result = response_dict.get(gas_price_strategy, None)
        if type(result) not in [int, float]:  # pragma: nocover
            raise ValueError(f"Invalid return value for `{gas_price_strategy}`!")
        gwei_result = result / 10  # adjustment (see api documentation)
        wei_result = web3.to_wei(gwei_result, "gwei")
        return {"gasPrice": wei_result}

    return gas_station_gas_price_strategy


class SignedTransactionTranslator:
    """Translator for SignedTransaction."""

    @staticmethod
    def to_dict(signed_transaction: SignedTransaction) -> Dict[str, Union[str, int]]:
        """Write SignedTransaction to dict."""
        signed_transaction_dict: Dict[str, Union[str, int]] = {
            "raw_transaction": cast(str, signed_transaction.rawTransaction.hex()),
            "hash": cast(str, signed_transaction.hash.hex()),
            "r": cast(int, signed_transaction.r),
            "s": cast(int, signed_transaction.s),
            "v": cast(int, signed_transaction.v),
        }
        return signed_transaction_dict

    @staticmethod
    def from_dict(signed_transaction_dict: JSONLike) -> SignedTransaction:
        """Get SignedTransaction from dict."""
        if (
            not isinstance(signed_transaction_dict, dict)
            and len(signed_transaction_dict) == 5
        ):
            raise ValueError(  # pragma: nocover
                f"Invalid for conversion. Found object: {signed_transaction_dict}."
            )
        signed_transaction = SignedTransaction(
            rawTransaction=HexBytes(
                cast(str, signed_transaction_dict["raw_transaction"])
            ),
            hash=HexBytes(cast(str, signed_transaction_dict["hash"])),
            r=cast(int, signed_transaction_dict["r"]),
            s=cast(int, signed_transaction_dict["s"]),
            v=cast(int, signed_transaction_dict["v"]),
        )
        return signed_transaction


class AttributeDictTranslator:
    """Translator for AttributeDict."""

    @classmethod
    def _remove_hexbytes(cls, value: Any) -> Any:
        """Process value to remove hexbytes."""
        if value is None:
            return value
        if isinstance(value, HexBytes):
            return value.hex()
        if isinstance(value, list):
            return cls._process_list(value, cls._remove_hexbytes)
        if type(value) in (bool, int, float, str, bytes):
            return value
        if isinstance(value, AttributeDict):
            return cls.to_dict(value)
        raise NotImplementedError(  # pragma: nocover
            f"Unknown type conversion. Found type: {type(value)}"
        )

    @classmethod
    def _add_hexbytes(cls, value: Any) -> Any:
        """Process value to add hexbytes."""
        if value is None:
            return value
        if isinstance(value, str):
            try:
                int(value, 16)
                return HexBytes(value)
            except Exception:  # pylint: disable=broad-except
                return value
        if isinstance(value, list):
            return cls._process_list(value, cls._add_hexbytes)
        if isinstance(value, dict):
            return cls.from_dict(value)
        if type(value) in (bool, int, float, bytes):
            return value
        raise NotImplementedError(  # pragma: nocover
            f"Unknown type conversion. Found type: {type(value)}"
        )

    @classmethod
    def _process_list(cls, li: list, callable_name: Callable) -> List:
        """Simplify a list with process value."""
        return [callable_name(el) for el in li]

    @classmethod
    def _valid_key(cls, key: Any) -> str:
        """Check validity of key."""
        if isinstance(key, str):
            return key
        raise ValueError("Key must be string.")  # pragma: nocover

    @classmethod
    def to_dict(cls, attr_dict: Union[AttributeDict, TxReceipt, TxData]) -> JSONLike:
        """Simplify to dict."""
        if not isinstance(attr_dict, AttributeDict):
            raise ValueError("No AttributeDict provided.")  # pragma: nocover
        result = {
            cls._valid_key(key): cls._remove_hexbytes(value)
            for key, value in attr_dict.items()
        }
        return result

    @classmethod
    def from_dict(cls, di: JSONLike) -> AttributeDict:
        """Get back attribute dict."""
        if not isinstance(di, dict):
            raise ValueError("No dict provided.")  # pragma: nocover
        processed_dict = {
            cls._valid_key(key): cls._add_hexbytes(value) for key, value in di.items()
        }
        return AttributeDict(processed_dict)


class EthereumCrypto(Crypto[LocalAccount]):
    """Class wrapping the Account Generation from Ethereum ledger."""

    identifier = _ETHEREUM

    def __init__(
        self,
        private_key_path: Optional[str] = None,
        password: Optional[str] = None,
        extra_entropy: Union[str, bytes, int] = "",
    ) -> None:
        """
        Instantiate an ethereum crypto object.

        :param private_key_path: the private key path of the agent
        :param password: the password to encrypt/decrypt the private key.
        :param extra_entropy: add extra randomness to whatever randomness your OS can provide
        """
        super().__init__(
            private_key_path=private_key_path,
            password=password,
            extra_entropy=extra_entropy,
        )

        bytes_representation = Web3.to_bytes(hexstr=self.entity.key.hex())
        self._public_key = str(keys.PrivateKey(bytes_representation).public_key)
        self._address = self.entity.address

    @property
    def private_key(self) -> str:
        """
        Return a private key.

        64 random hex characters (i.e. 32 bytes) + "0x" prefix.

        :return: a private key string in hex format
        """
        return self.entity.key.hex()

    @property
    def public_key(self) -> str:
        """
        Return a public key in hex format.

        128 hex characters (i.e. 64 bytes) + "0x" prefix.

        :return: a public key string in hex format
        """
        return self._public_key

    @property
    def address(self) -> str:
        """
        Return the address for the key pair.

        40 hex characters (i.e. 20 bytes) + "0x" prefix.

        :return: an address string in hex format
        """
        return self._address

    @classmethod
    def load_private_key_from_path(
        cls, file_name: str, password: Optional[str] = None
    ) -> LocalAccount:
        """
        Load a private key in hex format from a file.

        :param file_name: the path to the hex file.
        :param password: the password to encrypt/decrypt the private key.
        :return: the Entity.
        """
        private_key = cls.load(file_name, password)
        try:
            if not private_key.startswith("0x"):
                hex_to_bytes_for_key(private_key)
        except KeyIsIncorrect as e:
            if not password:
                raise KeyIsIncorrect(
                    f"Error on key `{file_name}` load! Try to specify `password`: Error: {repr(e)} "
                ) from e
            raise KeyIsIncorrect(
                f"Error on key `{file_name}` load! Wrong password?: Error: {repr(e)} "
            ) from e

        account = Account.from_key(  # pylint: disable=no-value-for-parameter
            private_key=private_key
        )
        return account

    def sign_message(self, message: bytes, is_deprecated_mode: bool = False) -> str:
        """
        Sign a message in bytes string form.

        :param message: the message to be signed
        :param is_deprecated_mode: if the deprecated signing is used
        :return: signature of the message in string form
        """
        if is_deprecated_mode and len(message) == 32:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                signature_dict = self.entity.signHash(message)
            signed_msg = signature_dict["signature"].hex()
        else:
            signable_message = encode_defunct(primitive=message)
            signature = self.entity.sign_message(signable_message=signable_message)
            signed_msg = signature["signature"].hex()
        return signed_msg

    def sign_transaction(self, transaction: JSONLike) -> JSONLike:
        """
        Sign a transaction in bytes string form.

        :param transaction: the transaction to be signed
        :return: signed transaction
        """
        signed_transaction = cast(Account, self.entity).sign_transaction(
            transaction_dict=transaction
        )
        signed_transaction_dict = SignedTransactionTranslator.to_dict(
            signed_transaction
        )
        return cast(JSONLike, signed_transaction_dict)

    @classmethod
    def generate_private_key(
        cls, extra_entropy: Union[str, bytes, int] = ""
    ) -> LocalAccount:
        """
        Generate a key pair for ethereum network.

        :param extra_entropy: add extra randomness to whatever randomness your OS can provide
        :return: account object
        """
        account = Account.create(
            extra_entropy
        )  # pylint: disable=no-value-for-parameter
        return account

    def encrypt(self, password: str) -> str:
        """
        Encrypt the private key and return in json.

        :param password: the password to decrypt.
        :return: json string containing encrypted private key.
        """
        encrypted = Account.encrypt(self.private_key, password)
        return json.dumps(encrypted)

    @classmethod
    def decrypt(cls, keyfile_json: str, password: str) -> str:
        """
        Decrypt the private key and return in raw form.

        :param keyfile_json: json str containing encrypted private key.
        :param password: the password to decrypt.
        :return: the raw private key (without leading "0x").
        """
        try:
            private_key = Account.decrypt(keyfile_json, password)
        except ValueError as e:
            if e.args[0] == "MAC mismatch":
                raise DecryptError() from e
            raise
        return private_key.hex()[2:]


class EthereumHelper(Helper):
    """Helper class usable as Mixin for EthereumApi or as standalone class."""

    @staticmethod
    def is_transaction_settled(tx_receipt: JSONLike) -> bool:
        """
        Check whether a transaction is settled or not.

        :param tx_receipt: the receipt associated to the transaction.
        :return: True if the transaction has been settled, False o/w.
        """
        is_successful = False
        if tx_receipt is not None:
            is_successful = tx_receipt.get("status", 0) == 1
        return is_successful

    @staticmethod
    def get_contract_address(tx_receipt: JSONLike) -> Optional[str]:
        """
        Retrieve the `contract_address` from a transaction receipt.

        :param tx_receipt: the receipt of the transaction.
        :return: the contract address, if present
        """
        contract_address = cast(Optional[str], tx_receipt.get("contractAddress", None))
        return contract_address

    @staticmethod
    def is_transaction_valid(
        tx: dict,
        seller: Address,
        client: Address,
        tx_nonce: str,
        amount: int,
    ) -> bool:
        """
        Check whether a transaction is valid or not.

        :param tx: the transaction.
        :param seller: the address of the seller.
        :param client: the address of the client.
        :param tx_nonce: the transaction nonce.
        :param amount: the amount we expect to get from the transaction.
        :return: True if the random_message is equals to tx['input']
        """
        is_valid = False
        if tx is not None:
            is_valid = (
                tx.get("input") == tx_nonce
                and tx.get("value") == amount
                and tx.get("from") == client
                and tx.get("to") == seller
            )
        return is_valid

    @staticmethod
    def generate_tx_nonce(seller: Address, client: Address) -> str:
        """
        Generate a unique hash to distinguish transactions with the same terms.

        :param seller: the address of the seller.
        :param client: the address of the client.
        :return: return the hash in hex.
        """
        aggregate_hash = Web3.keccak(
            b"".join([seller.encode(), client.encode(), uuid4().bytes])
        )
        return aggregate_hash.hex()

    @classmethod
    def get_address_from_public_key(cls, public_key: str) -> str:
        """
        Get the address from the public key.

        :param public_key: the public key
        :return: str
        """
        keccak_hash = Web3.keccak(hexstr=public_key)
        raw_address = keccak_hash[-20:].hex()
        address = Web3.to_checksum_address(raw_address)
        return address

    @classmethod
    def recover_message(
        cls, message: bytes, signature: str, is_deprecated_mode: bool = False
    ) -> Tuple[Address, ...]:
        """
        Recover the addresses from the hash.

        :param message: the message we expect
        :param signature: the transaction signature
        :param is_deprecated_mode: if the deprecated signing was used
        :return: the recovered addresses
        """
        if is_deprecated_mode:
            enforce(len(message) == 32, "Message must be hashed to exactly 32 bytes.")
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                address = Account._recover_hash(  # pylint: disable=no-value-for-parameter,protected-access
                    message_hash=message, signature=signature
                )
        else:
            signable_message = encode_defunct(primitive=message)
            address = Account.recover_message(  # pylint: disable=no-value-for-parameter
                signable_message=signable_message, signature=signature
            )
        return (address,)

    @classmethod
    def recover_public_keys_from_message(
        cls, message: bytes, signature: str, is_deprecated_mode: bool = False
    ) -> Tuple[str, ...]:
        """
        Get the public key used to produce the `signature` of the `message`

        :param message: raw bytes used to produce signature
        :param signature: signature of the message
        :param is_deprecated_mode: if the deprecated signing was used
        :return: the recovered public keys
        """
        if not is_deprecated_mode:
            signable_message = encode_defunct(primitive=message)
            message = _hash_eip191_message(signable_message)
        hash_bytes = HexBytes(message)
        # code taken from https://github.com/ethereum/eth-account/blob/master/eth_account/account.py#L428
        if len(hash_bytes) != 32:  # pragma: nocover
            raise ValueError("The message hash must be exactly 32-bytes")
        signature_bytes = HexBytes(signature)
        signature_bytes_standard = to_standard_signature_bytes(signature_bytes)
        signature_obj = keys.Signature(signature_bytes=signature_bytes_standard)
        pubkey = signature_obj.recover_public_key_from_msg_hash(hash_bytes)
        return (str(pubkey),)

    @staticmethod
    def get_hash(message: bytes) -> str:
        """
        Get the hash of a message.

        :param message: the message to be hashed.
        :return: the hash of the message as a hex string.
        """
        digest = Web3.keccak(message).hex()
        return digest

    @classmethod
    def load_contract_interface(cls, file_path: Path) -> Dict[str, str]:
        """
        Load contract interface.

        :param file_path: the file path to the interface
        :return: the interface
        """
        with open_file(file_path, "r") as interface_file_ethereum:
            contract_interface = json.load(interface_file_ethereum)
        for key in [_ABI, _BYTECODE]:
            if key not in contract_interface:  # pragma: nocover
                raise ValueError(f"Contract {file_path} missing key {key}.")
        return contract_interface


class EthereumApi(LedgerApi, EthereumHelper):
    """Class to interact with the Ethereum Web3 APIs."""

    identifier = _ETHEREUM

    _gas_price_strategy_callables: Dict[str, Callable] = {
        GAS_STATION: get_gas_price_strategy,
        EIP1559: get_gas_price_strategy_eip1559,
        EIP1559_POLYGON: get_gas_price_strategy_eip1559_polygon,
    }

    def __init__(self, **kwargs: Any):
        """
        Initialize the Ethereum ledger APIs.

        :param kwargs: keyword arguments
        """
        self._api = Web3(
            HTTPProvider(endpoint_uri=kwargs.pop("address", DEFAULT_ADDRESS))
        )
        self._chain_id = kwargs.pop("chain_id", DEFAULT_CHAIN_ID)
        self._is_gas_estimation_enabled = kwargs.pop("is_gas_estimation_enabled", False)

        self._default_gas_price_strategy: str = kwargs.pop(
            "default_gas_price_strategy", EIP1559
        )
        if self._default_gas_price_strategy not in AVAILABLE_STRATEGIES:
            raise ValueError(
                f"Gas price strategy must be one of {AVAILABLE_STRATEGIES}, provided: {self._default_gas_price_strategy}"
            )  # pragma: nocover

        self._gas_price_strategies: Dict[str, Dict] = kwargs.pop(
            "gas_price_strategies", DEFAULT_GAS_PRICE_STRATEGIES
        )

        self._poa_chain = kwargs.pop("poa_chain", False)
        if self._poa_chain:
            # https://web3py.readthedocs.io/en/stable/middleware.html#geth-style-proof-of-authority
            self._api.middleware_onion.inject(
                geth_poa_middleware, name="geth_poa_middleware", layer=0
            )
            _default_logger.info(
                "EthereumApi has been configured with Proof of Authority chain support"
            )

    @property
    def api(self) -> Web3:
        """Get the underlying API object."""
        return self._api

    def get_balance(
        self, address: Address, raise_on_try: bool = False
    ) -> Optional[int]:
        """Get the balance of a given account."""
        return self._try_get_balance(address, raise_on_try=raise_on_try)

    @try_decorator("Unable to retrieve balance: {}", logger_method="warning")
    def _try_get_balance(self, address: Address, **_kwargs: Any) -> Optional[int]:
        """Get the balance of a given account."""
        check_address = self._api.to_checksum_address(address)
        return self._api.eth.get_balance(check_address)  # pylint: disable=no-member

    def get_state(
        self, callable_name: str, *args: Any, raise_on_try: bool = False, **kwargs: Any
    ) -> Optional[JSONLike]:
        """Call a specified function on the ledger API."""
        response = self._try_get_state(
            callable_name, *args, raise_on_try=raise_on_try, **kwargs
        )
        return response

    @try_decorator("Unable to get state: {}", logger_method="warning")
    def _try_get_state(  # pylint: disable=unused-argument
        self, callable_name: str, *args: Any, **kwargs: Any
    ) -> Optional[JSONLike]:
        """Try to call a function on the ledger API."""

        if "raise_on_try" in kwargs:
            logging.info(
                f"popping `raise_on_try` from {self.__class__.__name__}.get_state kwargs"
            )
            kwargs.pop("raise_on_try")

        function = getattr(self._api.eth, callable_name)
        response = function(*args, **kwargs)

        if isinstance(response, AttributeDict):
            result = AttributeDictTranslator.to_dict(response)
            return result

        if type(response) in (int, float, bytes, str, list, dict):  # pragma: nocover
            # missing full checks for nested objects
            return {f"{callable_name}_result": response}
        raise NotImplementedError(  # pragma: nocover
            f"Response must be of types=int, float, bytes, str, list, dict. Found={type(response)}."
        )

    def get_transfer_transaction(  # pylint: disable=arguments-differ
        self,
        sender_address: Address,
        destination_address: Address,
        amount: int,
        tx_fee: int,
        tx_nonce: str,
        chain_id: Optional[int] = None,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[str] = None,
        gas_price: Optional[str] = None,
        gas_price_strategy: Optional[str] = None,
        gas_price_strategy_extra_config: Optional[Dict] = None,
        raise_on_try: bool = False,
        **kwargs: Any,
    ) -> Optional[JSONLike]:
        """
        Submit a transfer transaction to the ledger.

        :param sender_address: the sender address of the payer.
        :param destination_address: the destination address of the payee.
        :param amount: the amount of wealth to be transferred (in Wei).
        :param tx_fee: the transaction fee (gas) to be used (in Wei).
        :param tx_nonce: verifies the authenticity of the tx.
        :param chain_id: the Chain ID of the Ethereum transaction.
        :param max_fee_per_gas: maximum amount you’re willing to pay, inclusive of `baseFeePerGas` and `maxPriorityFeePerGas`. The difference between `maxFeePerGas` and `baseFeePerGas + maxPriorityFeePerGas` is refunded  (in Wei).
        :param max_priority_fee_per_gas: the part of the fee that goes to the miner (in Wei).
        :param gas_price: the gas price (in Wei)
        :param gas_price_strategy: the gas price strategy to be used.
        :param gas_price_strategy_extra_config: extra config for gas price strategy.
        :param raise_on_try: whether the method will raise or log on error
        :param kwargs: keyword arguments
        :return: the transfer transaction
        """
        transaction: Optional[JSONLike] = None
        chain_id = chain_id if chain_id is not None else self._chain_id
        destination_address = self._api.to_checksum_address(destination_address)
        sender_address = self._api.to_checksum_address(sender_address)
        nonce = self._try_get_transaction_count(
            sender_address,
            raise_on_try=raise_on_try,
        )
        if nonce is None:
            return transaction
        transaction = {
            "nonce": nonce,
            "chainId": chain_id,
            "to": destination_address,
            "value": amount,
            "gas": tx_fee,
            "data": tx_nonce,
        }
        if self._is_gas_estimation_enabled:
            transaction = self.update_with_gas_estimate(transaction)

        if max_fee_per_gas is not None:
            max_priority_fee_per_gas = (
                self._try_get_max_priority_fee(raise_on_try=raise_on_try)
                if max_priority_fee_per_gas is None
                else max_priority_fee_per_gas
            )
            transaction.update(
                {
                    "maxFeePerGas": max_fee_per_gas,
                    "maxPriorityFeePerGas": max_priority_fee_per_gas,
                }
            )

        if gas_price is not None:
            transaction.update({"gasPrice": gas_price})

        if gas_price is None and max_fee_per_gas is None:
            gas_pricing = self.try_get_gas_pricing(
                gas_price_strategy,
                gas_price_strategy_extra_config,
                raise_on_try=raise_on_try,
            )
            if gas_pricing is None:
                return transaction  # pragma: nocover
            transaction.update(gas_pricing)

        return transaction

    def _get_gas_price_strategy(
        self,
        gas_price_strategy: Optional[str] = None,
        extra_config: Optional[Dict] = None,
    ) -> Optional[Tuple[str, Callable]]:
        """
        Returns parameters for gas price callable.

        Note: The priority of gas price callable will be
        `extra_config(Runtime params) > self._gas_price_strategies (Set using config file.) > DEFAULT_GAS_PRICE_STRATEGIES (Default values.)`

        :param gas_price_strategy: name of the gas price strategy.
        :param extra_config: gas price strategy getter parameters.
        :return: gas price strategy's name and callable.
        """
        gas_price_strategy = (
            gas_price_strategy
            if gas_price_strategy is not None
            else self._default_gas_price_strategy
        )
        if gas_price_strategy not in AVAILABLE_STRATEGIES:  # pragma: nocover
            _default_logger.debug(
                f"Gas price strategy must be one of {AVAILABLE_STRATEGIES}, provided: {self._default_gas_price_strategy}"
            )
            return None

        _default_logger.debug(f"Using strategy: {gas_price_strategy}")
        gas_price_strategy_getter = self._gas_price_strategy_callables[
            gas_price_strategy
        ]

        parameters = cast(dict, DEFAULT_GAS_PRICE_STRATEGIES.get(gas_price_strategy))
        parameters.update(self._gas_price_strategies.get(gas_price_strategy, {}))
        parameters.update(extra_config or {})
        return gas_price_strategy, gas_price_strategy_getter(**parameters)

    @staticmethod
    def __reprice(old_price: Wei) -> Wei:
        return Wei(math.ceil(old_price * TIP_INCREASE))

    @try_decorator("Unable to retrieve gas price: {}", logger_method="warning")
    def try_get_gas_pricing(
        self,
        gas_price_strategy: Optional[str] = None,
        extra_config: Optional[Dict] = None,
        old_price: Optional[Dict[str, Wei]] = None,
        **_kwargs: Any,
    ) -> Optional[Dict[str, Wei]]:
        """
        Try get the gas price based on the provided strategy.

        :param gas_price_strategy: the gas price strategy to use, e.g., the EIP-1559 strategy.
            Can be either `eip1559` or `gas_station`.
        :param extra_config: gas price strategy getter parameters.
        :param old_price: the old gas price params in case that we are trying to resubmit a transaction.
        :param _kwargs: the keyword arguments. Possible kwargs are:
            `raise_on_try`: bool flag specifying whether the method will raise or log on error (used by `try_decorator`)
        :return: a dictionary with the gas data.
        """

        retrieved_strategy = self._get_gas_price_strategy(
            gas_price_strategy,
            extra_config,
        )
        if retrieved_strategy is None:  # pragma: nocover
            return None
        gas_price_strategy, gas_price_strategy_callable = retrieved_strategy

        prior_strategy = (
            self._api.eth._gas_price_strategy  # pylint: disable=protected-access
        )
        try:
            self._api.eth.set_gas_price_strategy(gas_price_strategy_callable)
            gas_price = self._api.eth.generate_gas_price()
        finally:
            self._api.eth.set_gas_price_strategy(prior_strategy)  # pragma: nocover

        if gas_price is None or old_price is None:
            return gas_price

        gas_price = cast(Dict[str, Wei], gas_price)
        if gas_price_strategy in (EIP1559, EIP1559_POLYGON):
            updated_max_fee_per_gas = self.__reprice(old_price["maxFeePerGas"])
            updated_max_priority_fee_per_gas = self.__reprice(
                old_price["maxPriorityFeePerGas"]
            )

            if gas_price["maxFeePerGas"] < updated_max_fee_per_gas:
                gas_price["maxFeePerGas"] = updated_max_fee_per_gas
                gas_price["maxPriorityFeePerGas"] = updated_max_priority_fee_per_gas

        elif gas_price_strategy == GAS_STATION:
            updated_gas_price = self.__reprice(old_price["gasPrice"])
            gas_price["gasPrice"] = max(gas_price["gasPrice"], updated_gas_price)

        return gas_price

    @try_decorator("Unable to retrieve transaction count: {}", logger_method="warning")
    def _try_get_transaction_count(
        self, address: Address, **_kwargs: Any
    ) -> Optional[int]:
        """Try get the transaction count."""
        nonce = self._api.eth.get_transaction_count(  # pylint: disable=no-member
            self._api.to_checksum_address(address)
        )
        return nonce

    def update_with_gas_estimate(  # pylint: disable=arguments-differ
        self, transaction: JSONLike, raise_on_try: bool = False
    ) -> JSONLike:
        """
        Attempts to update the transaction with a gas estimate

        :param transaction: the transaction
        :param raise_on_try: whether the method will raise or log on error
        :return: the updated transaction
        """
        gas_estimate = self._try_get_gas_estimate(
            transaction,
            raise_on_try=raise_on_try,
        )
        if gas_estimate is not None:
            transaction["gas"] = gas_estimate
        return transaction

    @try_decorator("Unable to retrieve gas estimate: {}", logger_method="warning")
    def _try_get_gas_estimate(
        self, transaction: JSONLike, **_kwargs: Any
    ) -> Optional[int]:
        """Try get the gas estimate."""
        gas_estimate: Optional[int] = None
        transaction = deepcopy(transaction)
        del transaction["gas"]
        try:
            gas_estimate = self._api.eth.estimate_gas(  # pylint: disable=no-member
                transaction=cast(
                    TxParams, AttributeDictTranslator.from_dict(transaction)
                )
            )
        except (ContractLogicError, ValueError) as e:
            _default_logger.warning(
                f"Unable to estimate gas with default state , "
                f"{type(e).__name__}: {e.__str__()}"
            )
            # gas estimation might fail when repricing txs
            # to avoid effects of pending txs when estimating gas
            # we can set the block identifier to "latest" block
            # this might fail if the node doesn't support the `block_identifier` param
            gas_estimate = self._api.eth.estimate_gas(  # pylint: disable=no-member
                transaction=cast(
                    TxParams, AttributeDictTranslator.from_dict(transaction)
                ),
                block_identifier="latest",
            )

        return gas_estimate

    def send_signed_transaction(
        self, tx_signed: JSONLike, raise_on_try: bool = False
    ) -> Optional[str]:
        """
        Send a signed transaction and wait for confirmation.

        :param tx_signed: the signed transaction
        :param raise_on_try: whether the method will raise or log on error
        :return: tx_digest, if present
        """
        tx_digest = self._try_send_signed_transaction(
            tx_signed, raise_on_try=raise_on_try
        )
        return tx_digest

    @try_decorator("Unable to send transaction: {}", logger_method="warning")
    def _try_send_signed_transaction(
        self, tx_signed: JSONLike, **_kwargs: Any
    ) -> Optional[str]:
        """
        Try send a signed transaction.

        :param tx_signed: the signed transaction
        :param _kwargs: the keyword arguments. Possible kwargs are:
            `raise_on_try`: bool flag specifying whether the method will raise or log on error (used by `try_decorator`)
        :return: tx_digest, if present
        """
        signed_transaction = SignedTransactionTranslator.from_dict(tx_signed)
        hex_value = self._api.eth.send_raw_transaction(  # pylint: disable=no-member
            signed_transaction.rawTransaction
        )
        tx_digest = hex_value.hex()
        _default_logger.debug(
            "Successfully sent transaction with digest: {}".format(tx_digest)
        )
        return tx_digest

    def get_transaction_receipt(
        self, tx_digest: str, raise_on_try: bool = False
    ) -> Optional[JSONLike]:
        """
        Get the transaction receipt for a transaction digest.

        :param tx_digest: the digest associated to the transaction.
        :param raise_on_try: whether the method will raise or log on error
        :return: the tx receipt, if present
        """
        tx_receipt = self._try_get_transaction_receipt(
            tx_digest,
            raise_on_try=raise_on_try,
        )

        if tx_receipt is not None and not bool(tx_receipt["status"]):
            tx = self.get_transaction(tx_digest, raise_on_try=raise_on_try)
            tx_receipt["revert_reason"] = self._try_get_revert_reason(
                tx,
                raise_on_try=raise_on_try,
            )

        return tx_receipt

    @try_decorator(
        "Error when attempting getting tx receipt: {}", logger_method="debug"
    )
    def _try_get_transaction_receipt(
        self, tx_digest: str, **_kwargs: Any
    ) -> Optional[JSONLike]:
        """
        Try get the transaction receipt.

        :param tx_digest: the digest associated to the transaction.
        :param _kwargs: the keyword arguments. Possible kwargs are:
            `raise_on_try`: bool flag specifying whether the method will raise or log on error (used by `try_decorator`)
        :return: the tx receipt, if present
        """
        tx_receipt = self._api.eth.get_transaction_receipt(  # pylint: disable=no-member
            cast(HexStr, tx_digest)
        )
        return AttributeDictTranslator.to_dict(tx_receipt)

    def get_transaction(
        self,
        tx_digest: str,
        raise_on_try: bool = False,
    ) -> Optional[JSONLike]:
        """
        Get the transaction for a transaction digest.

        :param tx_digest: the digest associated to the transaction.
        :param raise_on_try: whether the method will raise or log on error
        :return: the tx, if present
        """
        tx = self._try_get_transaction(tx_digest, raise_on_try=raise_on_try)
        return tx

    @try_decorator("Error when attempting getting tx: {}", logger_method="debug")
    def _try_get_transaction(
        self, tx_digest: str, **_kwargs: Any
    ) -> Optional[JSONLike]:
        """
        Get the transaction.

        :param tx_digest: the transaction digest.
        :param _kwargs: the keyword arguments. Possible kwargs are:
            `raise_on_try`: bool flag specifying whether the method will raise or log on error (used by `try_decorator`)
        :return: the tx, if found
        """
        tx = self._api.eth.get_transaction(
            cast(HexStr, tx_digest)
        )  # pylint: disable=no-member
        return AttributeDictTranslator.to_dict(tx)

    @try_decorator(
        "Error when attempting getting tx revert reason: {}", logger_method="debug"
    )
    def _try_get_revert_reason(self, tx: TxData, **_kwargs: Any) -> str:
        """Try to check the revert reason of a transaction.

        :param tx: the transaction for which we want to get the revert reason.
        :param _kwargs: the keyword arguments. Possible kwargs are:
            `raise_on_try`: bool flag specifying whether the method will raise or log on error (used by `try_decorator`)
        :return: the revert reason message.
        """
        # build a new transaction to replay:
        replay_tx = {
            "to": tx["to"],
            "from": tx["from"],
            "value": tx["value"],
            "data": tx["input"],
        }

        try:
            # replay the transaction on the provider
            self.api.eth.call(replay_tx, tx["blockNumber"] - 1)
        except ContractLogicError as e:
            # execution reverted exception
            return str(e)
        except HTTPError as e:
            # http exception
            raise e
        else:
            # given tx not reverted
            raise ValueError(f"The given transaction has not been reverted!\ntx: {tx}")

    def get_contract_instance(
        self, contract_interface: Dict[str, str], contract_address: Optional[str] = None
    ) -> Any:
        """
        Get the instance of a contract.

        :param contract_interface: the contract interface.
        :param contract_address: the contract address.
        :return: the contract instance
        """
        if contract_address is None:
            instance = self.api.eth.contract(
                abi=contract_interface[_ABI],
                bytecode=contract_interface[_BYTECODE],
            )
        else:
            _contract_address = self.api.to_checksum_address(contract_address)
            instance = self.api.eth.contract(
                address=_contract_address,
                abi=contract_interface[_ABI],
                bytecode=contract_interface[_BYTECODE],
            )
        return instance

    def get_deploy_transaction(  # pylint: disable=arguments-differ
        self,
        contract_interface: Dict[str, str],
        deployer_address: Address,
        raise_on_try: bool = False,
        **kwargs: Any,
    ) -> Optional[JSONLike]:
        """
        Get the transaction to deploy the smart contract.

        :param contract_interface: the contract interface.
        :param deployer_address: The address that will deploy the contract.
        :param raise_on_try: whether the method will raise or log on error
        :param kwargs: keyword arguments
        :return: the transaction dictionary.
        """

        # value to send to contract (in Wei)
        value: int = kwargs.pop("value", 0)

        # the gas to be used (in Wei)
        gas: Optional[int] = kwargs.pop("gas", None)

        # maximum amount you’re willing to pay, inclusive of `baseFeePerGas` and
        # `maxPriorityFeePerGas`. The difference between `maxFeePerGas` and
        # `baseFeePerGas + maxPriorityFeePerGas` is refunded  (in Wei).
        max_fee_per_gas: Optional[int] = kwargs.pop("max_fee_per_gas", None)

        # the part of the fee that goes to the miner (in Wei).
        max_priority_fee_per_gas: Optional[str] = kwargs.pop(
            "max_priority_fee_per_gas", None
        )

        # the gas price (in Wei)
        gas_price: Optional[str] = kwargs.pop("gas_price", None)

        # the gas price strategy to be used.
        gas_price_strategy: Optional[str] = kwargs.pop("gas_price_strategy", None)

        # extra config for gas price strategy.
        gas_price_strategy_extra_config: Optional[Dict] = kwargs.pop(
            "gas_price_strategy_extra_config", None
        )

        transaction: Optional[JSONLike] = None
        _deployer_address = self.api.to_checksum_address(deployer_address)
        nonce = self._try_get_transaction_count(
            _deployer_address, raise_on_try=raise_on_try
        )
        if nonce is None:
            return transaction
        instance = self.get_contract_instance(contract_interface)
        transaction = {
            "value": value,
            "nonce": nonce,
        }
        if max_fee_per_gas is not None:
            max_priority_fee_per_gas = (
                self._try_get_max_priority_fee(raise_on_try=raise_on_try)
                if max_priority_fee_per_gas is None
                else max_priority_fee_per_gas
            )
            if max_priority_fee_per_gas is None:
                return None  # pragma: nocover
            transaction.update(
                {
                    "maxFeePerGas": max_fee_per_gas,
                    "maxPriorityFeePerGas": max_priority_fee_per_gas,
                }
            )

        if gas_price is not None:
            transaction.update({"gasPrice": gas_price})

        if gas_price is None and max_fee_per_gas is None:
            gas_pricing = self.try_get_gas_pricing(
                gas_price_strategy,
                gas_price_strategy_extra_config,
                raise_on_try=raise_on_try,
            )

            if gas_pricing is None:
                return None  # pragma: nocover

            transaction.update(gas_pricing)

        transaction = instance.constructor(**kwargs).build_transaction(transaction)
        if transaction is None:
            return None  # pragma: nocover
        transaction.pop("to", None)  # only 'from' address, don't insert 'to' address!
        transaction.update({"from": _deployer_address})
        if gas is not None:
            transaction.update({"gas": gas})
        if self._is_gas_estimation_enabled:
            transaction = self.update_with_gas_estimate(transaction)
        return transaction

    @try_decorator("Unable to retrieve max_priority_fee: {}", logger_method="warning")
    def _try_get_max_priority_fee(self, **_kwargs: Any) -> str:
        """Try get the gas estimate."""
        return cast(str, self.api.eth.max_priority_fee)

    @classmethod
    def is_valid_address(cls, address: Address) -> bool:
        """
        Check if the address is valid.

        :param address: the address to validate
        :return: whether the address is valid
        """
        return Web3.is_address(address)

    @classmethod
    def contract_method_call(  # pylint: disable=arguments-differ
        cls,
        contract_instance: Any,
        method_name: str,
        **method_args: Any,
    ) -> Optional[JSONLike]:
        """Call a contract's method

        :param contract_instance: the contract to use
        :param method_name: the contract method to call
        :param method_args: the contract call parameters
        :return: the call result
        """
        method = getattr(contract_instance.functions, method_name)
        result = method(**method_args).call()
        return result

    def build_transaction(  # pylint: disable=too-many-arguments
        self,
        contract_instance: Any,
        method_name: str,
        method_args: Optional[Dict[Any, Any]],
        tx_args: Optional[Dict[Any, Any]],
        raise_on_try: bool = False,
    ) -> Optional[JSONLike]:
        """Prepare a transaction

        :param contract_instance: the contract to use
        :param method_name: the contract method to call
        :param method_args: the contract parameters
        :param tx_args: the transaction parameters
        :param raise_on_try: whether the method will raise or log on error
        :return: the transaction
        """

        if method_args is None:
            raise ValueError("Argument 'method_args' cannot be 'None'.")

        method = getattr(contract_instance.functions, method_name)
        tx = method(**cast(Dict, method_args))

        if tx_args is None:
            raise ValueError("Argument 'tx_args' cannot be 'None'.")

        tx_args = cast(Dict, tx_args)

        nonce = self.api.eth.get_transaction_count(tx_args["sender_address"])
        tx_params = {
            "nonce": nonce,
            "value": tx_args["value"] if "value" in tx_args else 0,
            "gas": 1,  # set this as a placeholder to avoid estimation on build_transaction()
        }

        # Parameter camel-casing due to contract api requirements
        for field in [
            "gas",
            "gasPrice",
            "maxFeePerGas",
            "maxPriorityFeePerGas",
        ]:
            if field in tx_args and tx_args[field] is not None:
                tx_params[field] = tx_args[field]

        if (
            "gasPrice" not in tx_params
            and "maxFeePerGas" not in tx_params
            and "maxPriorityFeePerGas" not in tx_params
        ):
            gas_data = self.try_get_gas_pricing(
                old_price=tx_args.get("old_price"), raise_on_try=raise_on_try
            )
            if gas_data:
                tx_params.update(gas_data)  # pragma: nocover

        tx = tx.build_transaction(tx_params)
        if self._is_gas_estimation_enabled:
            tx = self.update_with_gas_estimate(tx, raise_on_try=raise_on_try)

        return tx

    def get_transaction_transfer_logs(  # pylint: disable=too-many-arguments,too-many-locals
        self,
        contract_instance: Any,
        tx_hash: str,
        target_address: Optional[str] = None,
    ) -> Optional[JSONLike]:
        """
        Get all transfer events derived from a transaction.

        :param contract_instance: the contract
        :param tx_hash: the transaction hash
        :param target_address: optional address to filter transfer events to just those that affect it
        :return: the transfer logs
        """
        try:
            tx_receipt = self.api.eth.get_transaction_receipt(tx_hash)
            if tx_receipt is None:
                raise ValueError  # pragma: nocover

        except (TransactionNotFound, ValueError):  # pragma: nocover
            return dict(logs=[])

        transfer_logs = contract_instance.events.Transfer().processReceipt(tx_receipt)

        return dict(logs=transfer_logs)

    def send_signed_transactions(
        self,
        signed_transactions: List[JSONLike],
        raise_on_try: bool = False,
        **kwargs: Any,
    ) -> Optional[List[str]]:
        """
        Atomically send multiple of transactions.

        This operation is not supported for ethereum. Please use the ethereum_flashbots instead.

        :param signed_transactions: the signed transactions to bundle together and send.
        :param raise_on_try: whether the method will raise or log on error
        :param kwargs: the keyword arguments.
        """
        raise NotImplementedError(  # pragma: nocover
            f"Sending a bundle of transactions is not supported for the {self.identifier} plugin"
        )


class EthereumFaucetApi(FaucetApi):
    """Ethereum testnet faucet API."""

    identifier = _ETHEREUM
    testnet_name = TESTNET_NAME

    def get_wealth(self, address: Address, url: Optional[str] = None) -> None:
        """
        Get wealth from the faucet for the provided address.

        :param address: the address.
        :param url: the url
        """
        self._try_get_wealth(address, url)

    @staticmethod
    @try_decorator(
        "An error occured while attempting to generate wealth:\n{}",
        logger_method="error",
    )
    def _try_get_wealth(address: Address, url: Optional[str] = None) -> None:
        """
        Get wealth from the faucet for the provided address.

        :param address: the address.
        :param url: the url
        """
        if url is None:
            raise ValueError(  # pragma: nocover
                "Url is none, no default url provided. Please provide a faucet url."
            )
        response = requests.get(url + address)
        if response.status_code // 100 == 5:  # pragma: no cover
            _default_logger.error("Response: {}".format(response.status_code))
        elif response.status_code // 100 in [3, 4]:  # pragma: nocover
            response_dict = json.loads(response.text)
            _default_logger.warning(
                "Response: {}\nMessage: {}".format(
                    response.status_code, response_dict.get("message")
                )
            )
        elif response.status_code // 100 == 2:  # pragma: no cover
            response_dict = json.loads(response.text)
            _default_logger.info(
                "Response: {}\nMessage: {}".format(
                    response.status_code, response_dict.get("message")
                )
            )


class SimpleCacheLockWrapper:
    """Wrapper for session_cache with threading.Lock."""

    def __init__(self, session_cache: SimpleCache) -> None:
        """Init wrapper."""
        self.session_cache = session_cache
        self.lock = threading.Lock()

    def __contains__(self, *args: Any, **kwargs: Any) -> Any:
        """Contain item."""
        with self.lock:
            return self.session_cache.__contains__(*args, **kwargs)

    def __len__(self) -> int:
        """Length of the cache"""
        return len(self.session_cache)

    def cache(self, key: str, value: Any) -> Dict[str, Any]:
        """session_cache Cache."""
        return self.session_cache.cache(key=key, value=value)

    def get_cache_entry(self, key: str) -> Any:
        """Get cache entry."""
        with self.lock:
            return self.session_cache.get_cache_entry(key)

    def clear(self) -> None:
        """Clear cache entries."""
        with self.lock:
            self.session_cache.clear()

    def items(self) -> Dict[str, Any]:
        """Return session items."""
        return self.session_cache.items()


def set_wrapper_for_web3py_session_cache() -> None:
    """Wrap web3py session cache with threading.Lock."""

    # pylint: disable=protected-access
    web3._utils.request._session_cache = SimpleCacheLockWrapper(
        web3._utils.request._session_cache
    )


set_wrapper_for_web3py_session_cache()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 fetchai
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test messages module for gym protocol."""

# pylint: disable=too-many-statements,too-many-locals,no-member,too-few-public-methods,redefined-builtin
from typing import List

from aea.test_tools.test_protocol import BaseProtocolMessagesTestCase

from packages.fetchai.protocols.gym.custom_types import AnyObject
from packages.fetchai.protocols.gym.message import GymMessage


class TestMessageGym(BaseProtocolMessagesTestCase):
    """Test for the 'gym' protocol message."""

    MESSAGE_CLASS = GymMessage

    def build_messages(self) -> List[GymMessage]:  # type: ignore[override]
        """Build the messages to be used for testing."""
        return [
            GymMessage(
                performative=GymMessage.Performative.ACT,
                action=AnyObject("some_info"),
                step_id=12,
            ),
            GymMessage(
                performative=GymMessage.Performative.PERCEPT,
                step_id=12,
                observation=AnyObject("some_info1"),
                reward=1.0,
                done=True,
                info=AnyObject("some_info"),
            ),
            GymMessage(
                performative=GymMessage.Performative.STATUS,
                content={"some str": "some str"},
            ),
            GymMessage(
                performative=GymMessage.Performative.RESET,
            ),
            GymMessage(
                performative=GymMessage.Performative.CLOSE,
            ),
        ]

    def build_inconsistent(self) -> List[GymMessage]:  # type: ignore[override]
        """Build inconsistent messages to be used for testing."""
        return [
            GymMessage(
                performative=GymMessage.Performative.ACT,
                # skip content: action
                step_id=12,
            ),
            GymMessage(
                performative=GymMessage.Performative.PERCEPT,
                # skip content: step_id
                observation=AnyObject("some_info"),
                reward=1.4,
                done=True,
                info=AnyObject("some_info"),
            ),
            GymMessage(
                performative=GymMessage.Performative.STATUS,
                # skip content: content
            ),
        ]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains tests for aea manager utils."""
import time
from random import randint
from tempfile import TemporaryDirectory

import pytest

from aea.manager.utils import run_in_venv


pytestmark = pytest.mark.skip(
    reason="Unknown issue: https://github.com/valory-xyz/open-aea/actions/runs/7162860465/job/19500538744?pr=698"
)

RETURN_VALUE = randint(0, 2000)  # nosec

DEFAULT_TIMEOUT = 20


def _process_long():
    time.sleep(70)


def _process_exception():
    raise ValueError("Expected")


def _process_return_value(value_to_return):
    return value_to_return


def test_process_run_in_venv_timeout_error():
    """Test timeout error raised for process running too long."""
    with TemporaryDirectory() as tmp_dir:
        with pytest.raises(TimeoutError):
            run_in_venv(tmp_dir, _process_long, timeout=DEFAULT_TIMEOUT)


def test_process_run_in_venv_raise_custom_exception():
    """Test process returns expcetion."""
    with TemporaryDirectory() as tmp_dir:
        with pytest.raises(Exception, match="Expected"):
            run_in_venv(tmp_dir, _process_exception, timeout=DEFAULT_TIMEOUT)


def test_process_run_in_venv_return_value():
    """Test process return value."""
    with TemporaryDirectory() as tmp_dir:
        ret_value = run_in_venv(
            tmp_dir, _process_return_value, DEFAULT_TIMEOUT, RETURN_VALUE
        )
        assert ret_value == RETURN_VALUE



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains tests for aea.contracts.base."""
import logging
import os
import platform
import sys
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import cast
from unittest import mock
from unittest.mock import MagicMock

import pytest
import web3
from aea_ledger_ethereum import EthereumCrypto
from aea_ledger_ethereum.test_tools.constants import ETHEREUM_TESTNET_CONFIG
from aea_ledger_fetchai import FetchAICrypto

from aea.cli.scaffold import add_contract_abi, scaffold_item
from aea.cli.utils.context import Context
from aea.configurations.base import ComponentType, ContractConfig
from aea.configurations.constants import (  # noqa: F401  # pylint: disable=unused-import
    CONTRACT,
    CONTRACTS,
)
from aea.configurations.loader import load_component_configuration
from aea.contracts import contract_registry
from aea.contracts.base import Contract
from aea.contracts.scaffold.contract import MyScaffoldContract
from aea.crypto.ledger_apis import ETHEREUM_DEFAULT_ADDRESS, FETCHAI_DEFAULT_ADDRESS
from aea.crypto.registries import crypto_registry, ledger_apis_registry
from aea.exceptions import AEAComponentLoadException
from aea.package_manager.v1 import PackageManagerV1

from tests.conftest import ROOT_DIR, make_uri


logger = logging.getLogger(__name__)


def test_from_dir():
    """Tests the from dir and from config methods."""
    contract = Contract.from_dir(
        os.path.join(ROOT_DIR, "tests", "data", "dummy_contract")
    )
    assert contract is not None
    assert contract.contract_interface is not None
    assert isinstance(contract.contract_interface, dict)


def test_from_config_and_registration():
    """Tests the from config method and contract registry registration."""

    directory = Path(ROOT_DIR, "tests", "data", "dummy_contract")
    configuration = load_component_configuration(ComponentType.CONTRACT, directory)
    configuration._directory = directory
    configuration = cast(ContractConfig, configuration)

    if str(configuration.public_id) in contract_registry.specs:
        contract_registry.specs.pop(str(configuration.public_id))

    contract = Contract.from_config(configuration)
    assert contract is not None
    assert contract.contract_interface is not None
    assert isinstance(contract.contract_interface, dict)
    assert contract.configuration == configuration
    assert contract.id == configuration.public_id

    # the contract is registered as side-effect
    assert str(contract.public_id) in contract_registry.specs

    try:
        contract_registry.specs.pop(str(configuration.public_id))
    except Exception as e:
        logger.exception(e)


def test_from_config_negative():
    """Tests the from config method raises."""

    directory = Path(ROOT_DIR, "tests", "data", "dummy_contract")
    configuration = load_component_configuration(ComponentType.CONTRACT, directory)
    configuration._directory = directory
    configuration = cast(ContractConfig, configuration)

    if str(configuration.public_id) in contract_registry.specs:
        contract_registry.specs.pop(str(configuration.public_id))

    configuration.class_name = "WrongName"
    with pytest.raises(AEAComponentLoadException):
        _ = Contract.from_config(configuration)

    try:
        contract_registry.specs.pop(str(configuration.public_id))
    except Exception as e:
        logger.exception(e)


def test_non_implemented_class_methods():
    """Tests the non implemented class methods."""
    with pytest.raises(NotImplementedError):
        Contract.get_raw_transaction("ledger_api", "contract_address")

    with pytest.raises(NotImplementedError):
        Contract.get_raw_message("ledger_api", "contract_address")

    with pytest.raises(NotImplementedError):
        Contract.get_state("ledger_api", "contract_address")


@pytest.fixture()
def dummy_contract(request):
    """Dummy contract fixture."""
    directory = Path(ROOT_DIR, "tests", "data", "dummy_contract")
    configuration = load_component_configuration(ComponentType.CONTRACT, directory)
    configuration._directory = directory
    configuration = cast(ContractConfig, configuration)

    if str(configuration.public_id) in contract_registry.specs:
        contract_registry.specs.pop(str(configuration.public_id))

    # load into sys modules and register into contract registry
    contract = Contract.from_config(configuration)
    yield contract
    contract_registry.specs.pop(str(configuration.public_id))


def test_get_instance_no_address_ethereum(dummy_contract):
    """Tests get instance method with no address for ethereum."""
    ledger_api = ledger_apis_registry.make(
        EthereumCrypto.identifier,
        address=ETHEREUM_DEFAULT_ADDRESS,
    )
    instance = dummy_contract.get_instance(ledger_api)
    assert type(instance) == web3._utils.datatypes.PropertyCheckingFactory


@pytest.mark.integration
@pytest.mark.ledger
def test_get_deploy_transaction_ethereum(
    dummy_contract, ganache_addr, ganache_port, ganache
):
    """Tests the deploy transaction classmethod for ethereum."""
    aea_ledger_ethereum = crypto_registry.make(EthereumCrypto.identifier)
    config = ETHEREUM_TESTNET_CONFIG
    config.update(dict(address=make_uri(ganache_addr, ganache_port)))
    ledger_api = ledger_apis_registry.make(EthereumCrypto.identifier, **config)
    deploy_tx = dummy_contract.get_deploy_transaction(
        ledger_api, aea_ledger_ethereum.address
    )
    assert deploy_tx is not None and len(deploy_tx) == 7
    assert all(
        key in ["from", "value", "gas", "gasPrice", "nonce", "data", "chainId"]
        for key in deploy_tx.keys()
    )


def test_get_instance_no_address_cosmwasm(dummy_contract):
    """Tests get instance method with no address for fetchai."""
    ledger_api = ledger_apis_registry.make(
        FetchAICrypto.identifier,
        address=FETCHAI_DEFAULT_ADDRESS,
    )
    instance = dummy_contract.get_instance(ledger_api)
    assert instance is None


@pytest.mark.skipif(
    condition=(
        platform.system() == "Darwin"
        and sys.version_info.major == 3
        and sys.version_info.minor == 10
    ),
    reason="Fails because of cosmpy dependency on macOS/py310",
)
def test_get_deploy_transaction_cosmwasm(dummy_contract):
    """Tests the deploy transaction classmethod for fetchai."""
    aea_ledger_fetchai = crypto_registry.make(FetchAICrypto.identifier)
    ledger_api = ledger_apis_registry.make(
        FetchAICrypto.identifier,
        address=FETCHAI_DEFAULT_ADDRESS,
    )
    deploy_tx = dummy_contract.get_deploy_transaction(
        ledger_api, aea_ledger_fetchai.address, account_number=1, sequence=0
    )
    assert deploy_tx is not None and len(deploy_tx) == 2
    assert all(key in ["tx", "sign_data"] for key in deploy_tx.keys())


def test_scaffold():
    """Test the scaffold contract can be loaded/instantiated."""
    scaffold = MyScaffoldContract("config")
    kwargs = {"key": "value"}
    with pytest.raises(NotImplementedError):
        scaffold.get_raw_transaction("ledger_api", "contract_address", **kwargs)
    with pytest.raises(NotImplementedError):
        scaffold.get_raw_message("ledger_api", "contract_address", **kwargs)
    with pytest.raises(NotImplementedError):
        scaffold.get_state("ledger_api", "contract_address", **kwargs)


def test_scaffolded_contract_method_call():
    """Tests a contract method call."""

    @dataclass
    class AgentConfig:
        author = "dummy_author"
        contracts = ()
        agent_name = "dummy_agent"

    with tempfile.TemporaryDirectory() as temp_dir:
        packages_dir = Path(temp_dir, "packages")
        packages_dir.mkdir()
        PackageManagerV1(path=packages_dir).dump()

        ctx = Context(cwd=temp_dir, verbosity="DEBUG", registry_path=str(packages_dir))
        ctx.agent_config = AgentConfig()
        ctx.agent_config.directory = packages_dir / "dummy_author"
        ctx.config["to_local_registry"] = True

        contract_name = "i_uniswap_v2erc20"
        contract_abi_path = Path("tests", "test_contracts", "IUniswapV2ERC20.json")

        # Scaffold a new contract
        scaffold_item(ctx, CONTRACT, contract_name)
        add_contract_abi(ctx, contract_name, contract_abi_path)

        # Load the new contract
        contract_path = Path(packages_dir, "dummy_author", CONTRACTS, contract_name)
        contract = Contract.from_dir(str(contract_path))
        ledger_api = ledger_apis_registry.make(
            EthereumCrypto.identifier,
            address=ETHEREUM_DEFAULT_ADDRESS,
        )

        # Call a contract method: allowance
        SPENDER_ADDRESS = "0x7A1236d5195e31f1F573AD618b2b6FEFC85C5Ce6"
        OWNER_ADDRESS = "0x7A1236d5195e31f1F573AD618b2b6FEFC85C5Ce6"

        with mock.patch("web3.contract.contract.ContractFunction.call", return_value=0):
            res = contract.contract_method_call(
                ledger_api=ledger_api,
                method_name="allowance",
                owner=OWNER_ADDRESS,
                spender=SPENDER_ADDRESS,
            )

        assert res == 0


def test_contract_method_call():
    """Tests a contract method call."""
    contract = Contract.from_dir(
        os.path.join(ROOT_DIR, "tests", "data", "dummy_contract")
    )
    ledger_api = ledger_apis_registry.make(
        FetchAICrypto.identifier,
        address=FETCHAI_DEFAULT_ADDRESS,
    )
    with pytest.raises(NotImplementedError):
        contract.contract_method_call(ledger_api, "dummy_method")


def test_contract_method_call_2():
    """Tests a contract method call."""
    contract = Contract.from_dir(
        os.path.join(ROOT_DIR, "tests", "data", "dummy_contract")
    )
    ledger_api = MagicMock()
    ledger_api.contract_method_call.return_value = {}
    result = contract.contract_method_call(ledger_api, "dummy_method")
    assert result == {}


def test_build_transaction():
    """Tests a transaction build."""
    contract = Contract.from_dir(
        os.path.join(ROOT_DIR, "tests", "data", "dummy_contract")
    )
    ledger_api = ledger_apis_registry.make(
        FetchAICrypto.identifier,
        address=FETCHAI_DEFAULT_ADDRESS,
    )
    with pytest.raises(NotImplementedError):
        contract.build_transaction(ledger_api, "dummy_method", {}, {})


def test_build_transaction_2():
    """Tests a contract method call."""
    contract = Contract.from_dir(
        os.path.join(ROOT_DIR, "tests", "data", "dummy_contract")
    )
    ledger_api = MagicMock()
    ledger_api.build_transaction.return_value = {}
    result = contract.build_transaction(ledger_api, "dummy_method", {}, {})
    assert result == {}


def test_default_method_call():
    """Tests a default method build."""
    dummy_address = "0x0000000000000000000000000000000000000000"

    contract = Contract.from_dir(
        os.path.join(ROOT_DIR, "tests", "data", "dummy_contract")
    )

    ledger_api = ledger_apis_registry.make(
        EthereumCrypto.identifier,
        address=ETHEREUM_DEFAULT_ADDRESS,
    )

    # Call a function present in the ABI but not in the contract package
    with mock.patch("web3.contract.contract.ContractFunction.call", return_value=0):
        result = contract.default_method_call(
            ledger_api=ledger_api,
            contract_address=dummy_address,
            method_name="getAddress",
            _addr=dummy_address,
        )

        assert result == 0


def test_get_transaction_transfer_logs():
    """Tests a transaction log retrieval."""
    contract = Contract.from_dir(
        os.path.join(ROOT_DIR, "tests", "data", "dummy_contract")
    )
    ledger_api = ledger_apis_registry.make(
        FetchAICrypto.identifier,
        address=FETCHAI_DEFAULT_ADDRESS,
    )
    with pytest.raises(NotImplementedError):
        contract.get_transaction_transfer_logs(ledger_api, "dummy_hash")


def test_get_transaction_transfer_logs_2():
    """Tests a transaction log retrieval."""
    contract = Contract.from_dir(
        os.path.join(ROOT_DIR, "tests", "data", "dummy_contract")
    )
    ledger_api = MagicMock()
    ledger_api.get_transaction_transfer_logs.return_value = {}
    contract.get_transaction_transfer_logs(ledger_api, "dummy_hash")


def test_get_method_data():
    """Tests get_method_data."""
    contract = Contract.from_dir(
        os.path.join(ROOT_DIR, "tests", "data", "dummy_contract")
    )
    dummy_address = "0x0000000000000000000000000000000000000000"

    contract_instance = MagicMock()
    method = MagicMock()
    method.abi = {
        "inputs": [{"name": "inputA"}, {"name": "inputB"}, {"name": "inputC"}]
    }
    contract_instance.get_function_by_name.return_value = method
    contract_instance.encodeABI.return_value = dummy_address
    ledger_api = MagicMock()
    ledger_api.get_contract_instance.return_value = contract_instance

    res = contract.get_method_data(
        ledger_api=ledger_api,
        contract_address=dummy_address,
        method_name="dummy_method_name",
        inputA=0,
        inputB=0,
        inputC=0,
    )
    assert res == {
        "data": b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    }


def test_get_method_data__key_error():
    """Tests get_method_data."""
    contract = Contract.from_dir(
        os.path.join(ROOT_DIR, "tests", "data", "dummy_contract")
    )
    dummy_address = "0x0000000000000000000000000000000000000000"

    contract_instance = MagicMock()
    method = MagicMock()
    method.abi = {
        "inputs": [{"name": "inputA"}, {"name": "inputB"}, {"name": "inputC"}]
    }
    contract_instance.get_function_by_name.return_value = method
    contract_instance.encodeABI.return_value = dummy_address
    ledger_api = MagicMock()
    ledger_api.get_contract_instance.return_value = contract_instance

    res = contract.get_method_data(
        ledger_api=ledger_api,
        contract_address=dummy_address,
        method_name="dummy_method_name",
        inputA=0,
        inputB=0,
        # missing inputC to provoke a KeyError exception
    )
    assert not res


def test_get_method_data__attribute_error():
    """Tests get_method_data."""
    contract = Contract.from_dir(
        os.path.join(ROOT_DIR, "tests", "data", "dummy_contract")
    )
    dummy_address = "0x0000000000000000000000000000000000000000"

    contract_instance = MagicMock()
    method = None  # provoke an attribute error when accessing method.abi
    contract_instance.get_function_by_name.return_value = method
    contract_instance.encodeABI.return_value = dummy_address
    ledger_api = MagicMock()
    ledger_api.get_contract_instance.return_value = contract_instance

    res = contract.get_method_data(
        ledger_api=ledger_api,
        contract_address=dummy_address,
        method_name="dummy_method_name",
    )
    assert not res


def test_get_method_data_type_error():
    """Tests get_method_data."""
    contract = Contract.from_dir(
        os.path.join(ROOT_DIR, "tests", "data", "dummy_contract")
    )
    dummy_address = "0x0000000000000000000000000000000000000000"

    def dummy_fun(fn_name, args):
        raise TypeError

    contract_instance = MagicMock()
    method = MagicMock()
    method.abi = {
        "inputs": [{"name": "inputA"}, {"name": "inputB"}, {"name": "inputC"}]
    }
    contract_instance.get_function_by_name.return_value = method
    contract_instance.encodeABI = dummy_fun
    ledger_api = MagicMock()
    ledger_api.get_contract_instance.return_value = contract_instance

    res = contract.get_method_data(
        ledger_api=ledger_api,
        contract_address=dummy_address,
        method_name="dummy_method_name",
        inputA=0,
        inputB=0,
        inputC=0,
    )
    assert not res



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""The tests module contains the tests of the gym example."""

import os
import shutil
import sys
import tempfile
from pathlib import Path
from unittest.mock import MagicMock

from aea.configurations.base import ConnectionConfig
from aea.identity.base import Identity

from packages.fetchai.connections.gym.connection import GymConnection

from tests.common.pexpect_popen import PexpectWrapper
from tests.conftest import ROOT_DIR, env_path_separator


DIRECTORIES = ["packages", "examples"]


class TestGymExt:
    """Test the gym example."""

    @classmethod
    def setup_class(cls):
        """Set up the test class."""
        cls.old_cwd = Path(os.getcwd())
        cls.t = Path(tempfile.mkdtemp())
        for directory in DIRECTORIES:
            dir_path = Path(directory)
            tmp_dir = cls.t / dir_path
            src_dir = cls.old_cwd / Path(ROOT_DIR, directory)
            shutil.copytree(str(src_dir), str(tmp_dir))
        os.chdir(Path(cls.t))

    def test_gym_ex(self):
        """Run the gym ex sequence."""
        try:
            env = os.environ.copy()
            env[
                "PYTHONPATH"
            ] = f"{self.t}{env_path_separator()}{env.get('PYTHONPATH', '')}"
            process = PexpectWrapper(  # nosec
                [
                    sys.executable,
                    str(Path("examples/gym_ex/train.py").resolve()),
                    "--nb-steps",
                    "50",
                ],
                env=env,
                maxread=1,
                encoding="utf-8",
                logfile=sys.stdout,
            )

            process.expect(["Step 50/50"], timeout=10)
            process.wait_to_complete(5)
            assert process.returncode == 0, "Test failed"
        finally:
            process.terminate()
            process.wait_to_complete(5)

    @classmethod
    def teardown_class(cls):
        """Teardown the test."""
        try:
            os.chdir(Path(cls.old_cwd))
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


def test_gym_env_load():
    """Load gym env from file."""
    try:
        curdir = os.getcwd()
        os.chdir(os.path.join(ROOT_DIR, "examples", "gym_ex"))
        gym_env_path = "gyms.env.BanditNArmedRandom"
        configuration = ConnectionConfig(
            connection_id=GymConnection.connection_id, env=gym_env_path
        )
        identity = Identity(
            "name", address="agent_address", public_key="agent_public_key"
        )
        gym_con = GymConnection(
            gym_env=None,
            identity=identity,
            configuration=configuration,
            data_dir=MagicMock(),
        )
        assert gym_con.channel.gym_env is not None
    finally:
        os.chdir(curdir)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 fetchai
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains fipa's message definition."""

# pylint: disable=too-many-statements,too-many-locals,no-member,too-few-public-methods,too-many-branches,not-an-iterable,unidiomatic-typecheck,unsubscriptable-object
import logging
from typing import Any, Dict, Set, Tuple, cast

from aea.configurations.base import PublicId
from aea.exceptions import AEAEnforceError, enforce
from aea.protocols.base import Message  # type: ignore

from packages.fetchai.protocols.fipa.custom_types import (
    Description as CustomDescription,
)
from packages.fetchai.protocols.fipa.custom_types import Query as CustomQuery


_default_logger = logging.getLogger("aea.packages.fetchai.protocols.fipa.message")

DEFAULT_BODY_SIZE = 4


class FipaMessage(Message):
    """A protocol for FIPA ACL."""

    protocol_id = PublicId.from_str("fetchai/fipa:1.0.0")
    protocol_specification_id = PublicId.from_str("fetchai/fipa:1.0.0")

    Description = CustomDescription

    Query = CustomQuery

    class Performative(Message.Performative):
        """Performatives for the fipa protocol."""

        ACCEPT = "accept"
        ACCEPT_W_INFORM = "accept_w_inform"
        CFP = "cfp"
        DECLINE = "decline"
        END = "end"
        INFORM = "inform"
        MATCH_ACCEPT = "match_accept"
        MATCH_ACCEPT_W_INFORM = "match_accept_w_inform"
        PROPOSE = "propose"

        def __str__(self) -> str:
            """Get the string representation."""
            return str(self.value)

    _performatives = {
        "accept",
        "accept_w_inform",
        "cfp",
        "decline",
        "end",
        "inform",
        "match_accept",
        "match_accept_w_inform",
        "propose",
    }
    __slots__: Tuple[str, ...] = tuple()

    class _SlotsCls:
        __slots__ = (
            "dialogue_reference",
            "info",
            "message_id",
            "performative",
            "proposal",
            "query",
            "target",
        )

    def __init__(
        self,
        performative: Performative,
        dialogue_reference: Tuple[str, str] = ("", ""),
        message_id: int = 1,
        target: int = 0,
        **kwargs: Any,
    ):
        """
        Initialise an instance of FipaMessage.

        :param message_id: the message id.
        :param dialogue_reference: the dialogue reference.
        :param target: the message target.
        :param performative: the message performative.
        :param **kwargs: extra options.
        """
        super().__init__(
            dialogue_reference=dialogue_reference,
            message_id=message_id,
            target=target,
            performative=FipaMessage.Performative(performative),
            **kwargs,
        )

    @property
    def valid_performatives(self) -> Set[str]:
        """Get valid performatives."""
        return self._performatives

    @property
    def dialogue_reference(self) -> Tuple[str, str]:
        """Get the dialogue_reference of the message."""
        enforce(self.is_set("dialogue_reference"), "dialogue_reference is not set.")
        return cast(Tuple[str, str], self.get("dialogue_reference"))

    @property
    def message_id(self) -> int:
        """Get the message_id of the message."""
        enforce(self.is_set("message_id"), "message_id is not set.")
        return cast(int, self.get("message_id"))

    @property
    def performative(self) -> Performative:  # type: ignore # noqa: F821
        """Get the performative of the message."""
        enforce(self.is_set("performative"), "performative is not set.")
        return cast(FipaMessage.Performative, self.get("performative"))

    @property
    def target(self) -> int:
        """Get the target of the message."""
        enforce(self.is_set("target"), "target is not set.")
        return cast(int, self.get("target"))

    @property
    def info(self) -> Dict[str, str]:
        """Get the 'info' content from the message."""
        enforce(self.is_set("info"), "'info' content is not set.")
        return cast(Dict[str, str], self.get("info"))

    @property
    def proposal(self) -> CustomDescription:
        """Get the 'proposal' content from the message."""
        enforce(self.is_set("proposal"), "'proposal' content is not set.")
        return cast(CustomDescription, self.get("proposal"))

    @property
    def query(self) -> CustomQuery:
        """Get the 'query' content from the message."""
        enforce(self.is_set("query"), "'query' content is not set.")
        return cast(CustomQuery, self.get("query"))

    def _is_consistent(self) -> bool:
        """Check that the message follows the fipa protocol."""
        try:
            enforce(
                isinstance(self.dialogue_reference, tuple),
                "Invalid type for 'dialogue_reference'. Expected 'tuple'. Found '{}'.".format(
                    type(self.dialogue_reference)
                ),
            )
            enforce(
                isinstance(self.dialogue_reference[0], str),
                "Invalid type for 'dialogue_reference[0]'. Expected 'str'. Found '{}'.".format(
                    type(self.dialogue_reference[0])
                ),
            )
            enforce(
                isinstance(self.dialogue_reference[1], str),
                "Invalid type for 'dialogue_reference[1]'. Expected 'str'. Found '{}'.".format(
                    type(self.dialogue_reference[1])
                ),
            )
            enforce(
                type(self.message_id) is int,
                "Invalid type for 'message_id'. Expected 'int'. Found '{}'.".format(
                    type(self.message_id)
                ),
            )
            enforce(
                type(self.target) is int,
                "Invalid type for 'target'. Expected 'int'. Found '{}'.".format(
                    type(self.target)
                ),
            )

            # Light Protocol Rule 2
            # Check correct performative
            enforce(
                isinstance(self.performative, FipaMessage.Performative),
                "Invalid 'performative'. Expected either of '{}'. Found '{}'.".format(
                    self.valid_performatives, self.performative
                ),
            )

            # Check correct contents
            actual_nb_of_contents = len(self._body) - DEFAULT_BODY_SIZE
            expected_nb_of_contents = 0
            if self.performative == FipaMessage.Performative.CFP:
                expected_nb_of_contents = 1
                enforce(
                    isinstance(self.query, CustomQuery),
                    "Invalid type for content 'query'. Expected 'Query'. Found '{}'.".format(
                        type(self.query)
                    ),
                )
            elif self.performative == FipaMessage.Performative.PROPOSE:
                expected_nb_of_contents = 1
                enforce(
                    isinstance(self.proposal, CustomDescription),
                    "Invalid type for content 'proposal'. Expected 'Description'. Found '{}'.".format(
                        type(self.proposal)
                    ),
                )
            elif self.performative == FipaMessage.Performative.ACCEPT_W_INFORM:
                expected_nb_of_contents = 1
                enforce(
                    isinstance(self.info, dict),
                    "Invalid type for content 'info'. Expected 'dict'. Found '{}'.".format(
                        type(self.info)
                    ),
                )
                for key_of_info, value_of_info in self.info.items():
                    enforce(
                        isinstance(key_of_info, str),
                        "Invalid type for dictionary keys in content 'info'. Expected 'str'. Found '{}'.".format(
                            type(key_of_info)
                        ),
                    )
                    enforce(
                        isinstance(value_of_info, str),
                        "Invalid type for dictionary values in content 'info'. Expected 'str'. Found '{}'.".format(
                            type(value_of_info)
                        ),
                    )
            elif self.performative == FipaMessage.Performative.MATCH_ACCEPT_W_INFORM:
                expected_nb_of_contents = 1
                enforce(
                    isinstance(self.info, dict),
                    "Invalid type for content 'info'. Expected 'dict'. Found '{}'.".format(
                        type(self.info)
                    ),
                )
                for key_of_info, value_of_info in self.info.items():
                    enforce(
                        isinstance(key_of_info, str),
                        "Invalid type for dictionary keys in content 'info'. Expected 'str'. Found '{}'.".format(
                            type(key_of_info)
                        ),
                    )
                    enforce(
                        isinstance(value_of_info, str),
                        "Invalid type for dictionary values in content 'info'. Expected 'str'. Found '{}'.".format(
                            type(value_of_info)
                        ),
                    )
            elif self.performative == FipaMessage.Performative.INFORM:
                expected_nb_of_contents = 1
                enforce(
                    isinstance(self.info, dict),
                    "Invalid type for content 'info'. Expected 'dict'. Found '{}'.".format(
                        type(self.info)
                    ),
                )
                for key_of_info, value_of_info in self.info.items():
                    enforce(
                        isinstance(key_of_info, str),
                        "Invalid type for dictionary keys in content 'info'. Expected 'str'. Found '{}'.".format(
                            type(key_of_info)
                        ),
                    )
                    enforce(
                        isinstance(value_of_info, str),
                        "Invalid type for dictionary values in content 'info'. Expected 'str'. Found '{}'.".format(
                            type(value_of_info)
                        ),
                    )
            elif self.performative == FipaMessage.Performative.ACCEPT:
                expected_nb_of_contents = 0
            elif self.performative == FipaMessage.Performative.DECLINE:
                expected_nb_of_contents = 0
            elif self.performative == FipaMessage.Performative.MATCH_ACCEPT:
                expected_nb_of_contents = 0
            elif self.performative == FipaMessage.Performative.END:
                expected_nb_of_contents = 0

            # Check correct content count
            enforce(
                expected_nb_of_contents == actual_nb_of_contents,
                "Incorrect number of contents. Expected {}. Found {}".format(
                    expected_nb_of_contents, actual_nb_of_contents
                ),
            )

            # Light Protocol Rule 3
            if self.message_id == 1:
                enforce(
                    self.target == 0,
                    "Invalid 'target'. Expected 0 (because 'message_id' is 1). Found {}.".format(
                        self.target
                    ),
                )
        except (AEAEnforceError, ValueError, KeyError) as e:
            _default_logger.error(str(e))
            return False

        return True



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for the helpers module of the gym skill."""
# pylint: skip-file

from typing import cast
from unittest.mock import patch

import pytest

from packages.fetchai.protocols.gym.message import GymMessage
from packages.fetchai.skills.gym.dialogues import GymDialogue
from packages.fetchai.skills.gym.tests.intermediate_class import GymTestCase


class TestProxyEnv(GymTestCase):
    """Test ProxyEnv of gym."""

    is_agent_to_agent_messages = False

    def test__init__(self):
        """Test the __init__ method of the ProxyEnv class."""
        assert self.proxy_env._is_rl_agent_trained is False
        assert self.proxy_env._step_count == 0
        assert self.proxy_env._active_dialogue is None

    def test_properties(self):
        """Test the properties of the ProxyEnv class."""
        assert self.proxy_env.gym_dialogues == self.gym_dialogues

        with pytest.raises(ValueError, match="GymDialogue not set yet."):
            assert self.proxy_env.active_gym_dialogue
        self.proxy_env._active_dialogue = self.dummy_gym_dialogue
        assert self.proxy_env.active_gym_dialogue == self.dummy_gym_dialogue

        assert self.proxy_env.queue == self.proxy_env._queue
        assert self.proxy_env.is_rl_agent_trained is False

    def test_step_i(self):
        """Test the step method of the ProxyEnv class where no exception."""
        # setup
        action = "some_action"
        gym_dialogue = cast(
            GymDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.gym_dialogues,
                messages=self.list_of_gym_messages[:2],
            ),
        )
        self.proxy_env._active_dialogue = gym_dialogue
        percept_msg = self.build_incoming_message(
            message_type=GymMessage,
            performative=GymMessage.Performative.PERCEPT,
            step_id=self.proxy_env._step_count + 1,
            observation=self.mocked_observation,
            reward=self.mocked_reward,
            done=True,
            info=self.mocked_info,
        )

        # operation
        with patch.object(
            self.proxy_env._queue, "get", return_value=percept_msg
        ) as mocked_q_get:
            (
                actual_observation,
                actual_reward,
                actual_done,
                actual_info,
            ) = self.proxy_env.step(action)

        # after

        # _encode_and_send_action
        self.assert_quantity_in_outbox(1)

        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=GymMessage,
            performative=GymMessage.Performative.ACT,
            to=gym_dialogue.dialogue_label.dialogue_opponent_addr,
            sender=str(self.skill.skill_context.skill_id),
            action=GymMessage.AnyObject(action),
            step_id=self.proxy_env._step_count,
        )
        assert has_attributes, error_str

        mocked_q_get.assert_called_with(block=True, timeout=None)

        assert actual_observation == self.mocked_observation.any
        assert actual_reward == self.mocked_reward
        assert actual_done is True
        assert actual_info == self.mocked_info.any

    def test_step_ii(self):
        """Test the step method of the ProxyEnv class where performative is NOT percept."""
        # setup
        action = "some_action"
        gym_dialogue = cast(
            GymDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.gym_dialogues,
                messages=self.list_of_gym_messages[:2],
            ),
        )
        self.proxy_env._active_dialogue = gym_dialogue
        invalid_percept_msg = self.build_incoming_message(
            message_type=GymMessage,
            performative=GymMessage.Performative.RESET,
        )

        # operation
        with patch.object(
            self.proxy_env._queue, "get", return_value=invalid_percept_msg
        ) as mocked_q_get:
            with pytest.raises(
                ValueError,
                match=f"Unexpected performative. Expected={GymMessage.Performative.PERCEPT} got={invalid_percept_msg.performative}",
            ):
                self.proxy_env.step(action)

        # # after

        # _encode_and_send_action
        self.assert_quantity_in_outbox(1)

        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=GymMessage,
            performative=GymMessage.Performative.ACT,
            to=gym_dialogue.dialogue_label.dialogue_opponent_addr,
            sender=str(self.skill.skill_context.skill_id),
            action=GymMessage.AnyObject(action),
            step_id=self.proxy_env._step_count,
        )
        assert has_attributes, error_str

        mocked_q_get.assert_called_with(block=True, timeout=None)

    def test_step_iii(self):
        """Test the step method of the ProxyEnv class where performative is NOT percept."""
        # setup
        action = "some_action"
        gym_dialogue = cast(
            GymDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.gym_dialogues,
                messages=self.list_of_gym_messages[:2],
            ),
        )
        self.proxy_env._active_dialogue = gym_dialogue
        invalid_percept_msg = self.build_incoming_message(
            message_type=GymMessage,
            performative=GymMessage.Performative.PERCEPT,
            step_id=self.proxy_env._step_count,
            observation=self.mocked_observation,
            reward=self.mocked_reward,
            done=True,
            info=self.mocked_info,
        )

        # operation
        with patch.object(
            self.proxy_env._queue, "get", return_value=invalid_percept_msg
        ) as mocked_q_get:
            with pytest.raises(
                ValueError,
                match=f"Unexpected step id! expected={self.proxy_env._step_count+1}, actual={self.proxy_env._step_count}",
            ):
                self.proxy_env.step(action)

        # # after

        # _encode_and_send_action
        self.assert_quantity_in_outbox(1)

        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=GymMessage,
            performative=GymMessage.Performative.ACT,
            to=gym_dialogue.dialogue_label.dialogue_opponent_addr,
            sender=str(self.skill.skill_context.skill_id),
            action=GymMessage.AnyObject(action),
            step_id=self.proxy_env._step_count,
        )
        assert has_attributes, error_str

        mocked_q_get.assert_called_with(block=True, timeout=None)

    def test_reset_i(self):
        """Test the reset method of the ProxyEnv class where no exception."""
        # setup
        status_msg = self.build_incoming_message(
            message_type=GymMessage,
            performative=GymMessage.Performative.STATUS,
            content=self.mocked_status_content,
        )

        # operation
        with patch.object(
            self.proxy_env._queue, "get", return_value=status_msg
        ) as mocked_q_get:
            self.proxy_env.reset()

        # after
        assert self.proxy_env._step_count == 0
        assert self.proxy_env._is_rl_agent_trained is False
        assert self.proxy_env._active_dialogue is not None

        self.assert_quantity_in_outbox(1)
        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=GymMessage,
            performative=GymMessage.Performative.RESET,
            to=self.proxy_env.gym_address,
            sender=str(self.skill.skill_context.skill_id),
        )
        assert has_attributes, error_str

        mocked_q_get.assert_called_with(block=True, timeout=None)

    def test_reset_ii(self):
        """Test the reset method of the ProxyEnv class where performative is NOT status."""
        # setup
        invalid_msg = self.build_incoming_message(
            message_type=GymMessage,
            performative=GymMessage.Performative.RESET,
        )

        # operation
        with patch.object(
            self.proxy_env._queue, "get", return_value=invalid_msg
        ) as mocked_q_get:
            with pytest.raises(
                ValueError,
                match=f"Unexpected performative. Expected={GymMessage.Performative.STATUS} got={invalid_msg.performative}",
            ):
                self.proxy_env.reset()

        # after
        assert self.proxy_env._step_count == 0
        assert self.proxy_env._is_rl_agent_trained is False
        assert self.proxy_env._active_dialogue is not None

        self.assert_quantity_in_outbox(1)
        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=GymMessage,
            performative=GymMessage.Performative.RESET,
            to=self.proxy_env.gym_address,
            sender=str(self.skill.skill_context.skill_id),
        )
        assert has_attributes, error_str

        mocked_q_get.assert_called_with(block=True, timeout=None)

    def test_close_i(self):
        """Test the close method of the ProxyEnv class."""
        # setup
        self.proxy_env._is_rl_agent_trained = True
        gym_dialogue = cast(
            GymDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.gym_dialogues,
                messages=self.list_of_gym_messages[:4],
            ),
        )
        self.proxy_env._active_dialogue = gym_dialogue

        # operation
        self.proxy_env.close()

        # after
        self.assert_quantity_in_outbox(1)

        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=GymMessage,
            performative=GymMessage.Performative.CLOSE,
            to=gym_dialogue.dialogue_label.dialogue_opponent_addr,
            sender=str(self.skill.skill_context.skill_id),
        )
        assert has_attributes, error_str



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the handler classes of the ml_train skill."""
from typing import Tuple

import numpy as np


def produce_data(batch_size) -> Tuple:
    """Produce the data."""
    from tensorflow import keras  # pylint: disable=import-outside-toplevel

    ((train_x, train_y), _) = keras.datasets.fashion_mnist.load_data()

    idx = np.arange(train_x.shape[0])
    mask = np.zeros_like(idx, dtype=bool)

    selected = np.random.choice(idx, batch_size, replace=False)
    mask[selected] = True

    x_sample = train_x[mask]
    y_sample = train_y[mask]
    return x_sample, y_sample



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Implementation of the 'aea init' subcommand."""

from pathlib import Path
from typing import Dict, Optional

import click

from aea import __version__
from aea.cli.login import do_login
from aea.cli.register import do_register
from aea.cli.registry.settings import (
    DEFAULT_IPFS_URL,
    DEFAULT_REGISTRY_CONFIG,
    REGISTRY_CONFIG_KEY,
    REGISTRY_LOCAL,
    REGISTRY_REMOTE,
    REMOTE_HTTP,
    REMOTE_IPFS,
)
from aea.cli.registry.utils import check_is_author_logged_in, is_auth_token_present
from aea.cli.utils.click_utils import registry_flag, remote_registry_flag
from aea.cli.utils.config import get_or_create_cli_config, update_cli_config
from aea.cli.utils.constants import AEA_LOGO, AUTHOR_KEY, CLI_CONFIG_PATH
from aea.cli.utils.context import Context
from aea.cli.utils.decorators import pass_ctx
from aea.cli.utils.package_utils import validate_author_name


@click.command()
@click.option("--author", type=str, required=False)
@click.option("--reset", is_flag=True, help="Reset the initialization.")
@click.option("--no-subscribe", is_flag=True, help="For developers subscription.")
@registry_flag(mark_default=True, default_registry=REGISTRY_LOCAL)
@remote_registry_flag(mark_default=True, default_registry=REMOTE_IPFS)
@click.option(
    "--ipfs-node", type=str, default=DEFAULT_IPFS_URL, help="Multiaddr for IPFS node."
)
@pass_ctx
def init(  # pylint: disable=unused-argument
    ctx: Context,
    author: str,
    reset: bool,
    no_subscribe: bool,
    registry: str,
    remote_registry: str,
    ipfs_node: Optional[str],
) -> None:
    """Initialize your AEA configurations."""
    do_init(
        author,
        reset,
        no_subscribe,
        registry,
        default_remote_registry=remote_registry,
        ipfs_node=ipfs_node,
    )


def do_init(
    author: str,
    reset: bool,
    no_subscribe: bool,
    registry_type: str = REGISTRY_LOCAL,
    default_remote_registry: str = REMOTE_IPFS,
    ipfs_node: Optional[str] = None,
) -> None:
    """
    Initialize your AEA configurations.

    :param author: str author username.
    :param reset: True, if resetting the author name
    :param no_subscribe: bool flag for developers subscription skip on register.
    :param registry_type: default registry type.
    :param default_remote_registry: Default remote registry
    :param ipfs_node: URL to ipfs node
    """

    if reset or not Path(CLI_CONFIG_PATH).exists():
        author = validate_author_name(author)
        update_cli_config({AUTHOR_KEY: author})

        if registry_type == REGISTRY_LOCAL:
            _registry_init_local(default_remote_registry)
        else:
            _registry_init_remote(
                default_remote_registry, author, no_subscribe, ipfs_node
            )

        config = get_or_create_cli_config()
        config.pop(REGISTRY_CONFIG_KEY, None)  # for security reasons
        success_msg = "AEA configurations successfully initialized: {}".format(config)
    else:
        config = get_or_create_cli_config()
        config.pop(REGISTRY_CONFIG_KEY, None)  # for security reasons
        success_msg = "AEA configurations already initialized: {}. To reset use '--reset'.".format(
            config
        )
    click.echo(AEA_LOGO + "v" + __version__ + "\n")
    click.echo(success_msg)


def _registry_init_local(default_remote_registry: str = REMOTE_IPFS) -> None:
    """Initialize ipfs local"""
    registry_config = _set_registries(REGISTRY_LOCAL, default_remote_registry)
    update_cli_config({REGISTRY_CONFIG_KEY: registry_config})


def _registry_init_remote(
    default_remote_registry: str,
    author: str,
    no_subscribe: bool,
    ipfs_node: Optional[str],
) -> None:
    """Initialize remote registry"""
    if default_remote_registry == REMOTE_IPFS:
        _registry_init_ipfs(ipfs_node)
    else:
        _registry_init_http(username=author, no_subscribe=no_subscribe)


def _registry_init_ipfs(ipfs_node: Optional[str]) -> None:
    """Initialize ipfs registry"""

    registry_config = _set_registries(REGISTRY_REMOTE, REMOTE_IPFS)
    registry_config["settings"][REGISTRY_REMOTE][REMOTE_IPFS]["ipfs_node"] = ipfs_node
    update_cli_config({REGISTRY_CONFIG_KEY: registry_config})


def _registry_init_http(username: str, no_subscribe: bool) -> None:
    """
    Create an author name on the registry.

    :param username: the user name
    :param no_subscribe: bool flag for developers subscription skip on register.
    """
    registry_config = _set_registries(REGISTRY_REMOTE, REMOTE_HTTP)
    update_cli_config({REGISTRY_CONFIG_KEY: registry_config})

    if username is not None and is_auth_token_present():
        check_is_author_logged_in(username)
    else:
        is_registered = click.confirm("Do you have a Registry account?")
        if is_registered:
            password = click.prompt("Password", type=str, hide_input=True)
            do_login(username, password)
        else:
            click.echo("Create a new account on the Registry now:")
            email = click.prompt("Email", type=str)
            password = click.prompt("Password", type=str, hide_input=True)

            password_confirmation = ""  # nosec
            while password_confirmation != password:
                click.echo("Please make sure that passwords are equal.")
                password_confirmation = click.prompt(
                    "Confirm password", type=str, hide_input=True
                )

            do_register(username, email, password, password_confirmation, no_subscribe)


def _set_registries(
    default_registry: str = REGISTRY_LOCAL, default_remote_registry: str = REMOTE_IPFS
) -> Dict:
    """Set registry values."""
    registry_config = DEFAULT_REGISTRY_CONFIG.copy()
    registry_config["default"] = default_registry
    registry_config["settings"][REGISTRY_REMOTE]["default"] = default_remote_registry

    return registry_config



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains tests for P2PLibp2p connection."""

# pylint: skip-file

import pytest

from aea.test_tools.utils import wait_for_condition

from packages.valory.connections.p2p_libp2p.tests.base import libp2p_log_on_failure_all
from packages.valory.connections.test_libp2p.tests.base import (
    BaseP2PLibp2pTest,
    _make_libp2p_connection,
)


@libp2p_log_on_failure_all
class TestSlowQueue(BaseP2PLibp2pTest):
    """Test that libp2p node uses slow queue in case of long DHT lookups."""

    @classmethod
    def setup_class(cls):
        """Set the test up"""
        super().setup_class()

        connection_genesis = cls.make_connection()
        genesis_peer = connection_genesis.node.multiaddrs[0]

        cls.conn = cls.make_connection(entry_peers=[genesis_peer])
        for _ in range(2):
            cls.make_connection(entry_peers=[genesis_peer])

    def test_connection_is_established(self):
        """Test connection established."""
        assert self.all_connected

    @pytest.mark.asyncio
    async def test_slow_queue(self):
        """Test slow queue."""

        bad_address = _make_libp2p_connection().address
        good_address = self.multiplexers[-1].connections[-1].address

        for to in [good_address, bad_address]:
            sender = self.conn.address
            envelope = self.enveloped_default_message(to=to, sender=sender)
            await self.conn._node_client.send_envelope(envelope)

        def _check():
            with open(self.conn.node.log_file) as f:
                return "while sending slow envelope:" in f.read()

        wait_for_condition(_check, timeout=30, period=1)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Helpful utilities."""

import collections
import os
import shutil
import time
from contextlib import ExitStack, contextmanager
from copy import deepcopy
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    Mapping,
    MutableMapping,
    MutableSequence,
    MutableSet,
    Type,
    Union,
)


FULL_PERMISSION = 0o40777


def wait_for_condition(
    condition_checker: Callable,
    timeout: int = 2,
    error_msg: str = "Timeout",
    period: float = 0.001,
) -> None:
    """Wait for condition to occur in selected timeout."""

    start_time = time.time()
    while not condition_checker():
        time.sleep(period)
        if time.time() > start_time + timeout:
            raise TimeoutError(error_msg)


def consume(iterator: Iterable) -> None:
    """Consume the iterator"""
    collections.deque(iterator, maxlen=0)


@contextmanager
def as_context(*contexts: Any) -> Generator[None, None, None]:
    """Set contexts"""
    with ExitStack() as stack:
        consume(map(stack.enter_context, contexts))
        yield


def copy_class(cls: Type) -> Type:
    """Copy a class. Useful for testing class setup configurations"""

    # NOTE: this does not recursively deepcopy the class
    #       nor can it copy items that cannot be pickled,
    #       e.g. classes containing classmethods in __dict__.
    #       Use at your own risk.

    def is_mutable(obj: Any) -> bool:
        return isinstance(obj, (MutableSequence, MutableSet, MutableMapping))

    def deepcopy_if_mutable(mapping: Mapping[str, Any]) -> Dict[str, Any]:
        return {k: deepcopy(v) if is_mutable(v) else v for k, v in mapping.items()}

    return type(
        f"CopyOf{cls.__name__}",
        (cls, *cls.__bases__),
        deepcopy_if_mutable(cls.__dict__),
    )


def remove_test_directory(directory: Union[str, Path], retries: int = 3) -> bool:
    """Destroy a directory once tests are done, change permissions if needed.

    Note that on Windows directories and files that are open cannot be deleted.

    :param directory: directory to be deleted
    :param retries: number of re-attempts
    :return: whether the directory was successfully deleted
    """

    permission = os.stat(directory).st_mode
    while os.path.exists(directory) and retries:
        try:
            os.chmod(directory, FULL_PERMISSION)  # nosec
            shutil.rmtree(directory)
        except Exception:  # pylint: disable=broad-except
            retries -= 1
            time.sleep(1)
        finally:
            if os.path.exists(directory):
                os.chmod(directory, permission)
    return not os.path.exists(directory)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests for the identity module."""

import pytest
from aea_ledger_fetchai import FetchAICrypto

from aea.configurations.constants import DEFAULT_LEDGER
from aea.exceptions import AEAEnforceError
from aea.identity.base import Identity


def test_init_identity_positive():
    """Test initialization of the identity object."""
    assert Identity("some_name", address="some_address", public_key="some_public_key")
    assert Identity(
        "some_name",
        address="some_address",
        public_key="some_public_key",
        default_address_key=DEFAULT_LEDGER,
    )
    assert Identity(
        "some_name",
        addresses={
            DEFAULT_LEDGER: "some_address",
            FetchAICrypto.identifier: "some_address",
        },
        public_keys={
            DEFAULT_LEDGER: "some_public_key",
            FetchAICrypto.identifier: "some_public_key",
        },
    )
    assert Identity(
        "some_name",
        addresses={
            DEFAULT_LEDGER: "some_address",
            FetchAICrypto.identifier: "some_address",
        },
        public_keys={
            DEFAULT_LEDGER: "some_public_key",
            FetchAICrypto.identifier: "some_public_key",
        },
        default_address_key=DEFAULT_LEDGER,
    )


def test_init_identity_negative():
    """Test initialization of the identity object."""
    name = "some_name"
    address_1 = "some_address"
    addresses_1 = {"some_ledger_id": "some_address"}
    addresses_2 = {}
    public_key_1 = "some_public_key"
    public_keys_1 = {"some_ledger_id": "some_public_key"}
    public_keys_2 = {}
    with pytest.raises(ValueError, match="Provide a key for the default address."):
        Identity(name, default_address_key=None)
    with pytest.raises(
        ValueError,
        match="Either provide a single address or a dictionary of addresses, and not both.",
    ):
        Identity(name)
    with pytest.raises(
        ValueError,
        match="Either provide a single address or a dictionary of addresses, and not both.",
    ):
        Identity(name, address=address_1, addresses=addresses_1)
    with pytest.raises(ValueError, match="Provide at least one pair of addresses."):
        Identity(name, addresses=addresses_2)
    with pytest.raises(
        ValueError,
        match="If you provide a dictionary of addresses, you must provide its corresponding dictionary of public keys.",
    ):
        Identity(name, addresses=addresses_1)
    with pytest.raises(
        ValueError,
        match="If you provide a dictionary of addresses, you must not provide a single public key.",
    ):
        Identity(name, addresses=addresses_1, public_key=public_key_1)
    with pytest.raises(
        AEAEnforceError,
        match="Keys in public keys and addresses dictionaries do not match. They must be identical.",
    ):
        Identity(name, addresses=addresses_1, public_keys=public_keys_2)
    with pytest.raises(
        AEAEnforceError,
        match="The default address key must exist in both addresses and public keys dictionaries.",
    ):
        Identity(
            name,
            addresses=addresses_1,
            public_keys=public_keys_1,
            default_address_key="some_other_ledger",
        )
    with pytest.raises(
        ValueError,
        match="If you provide a single address, you must not provide a dictionary of public keys.",
    ):
        Identity(name, address=address_1, public_keys=public_keys_1)
    with pytest.raises(
        ValueError,
        match="If you provide a single address, you must provide its corresponding public key.",
    ):
        Identity(name, address=address_1)


def test_accessors():
    """Test the properties of the identity object."""
    name = "some_name"
    address = "some_address"
    public_key = "some_public_key"
    identity = Identity(name, address=address, public_key=public_key)
    assert identity.name == name
    assert identity.address == address
    assert identity.addresses == {DEFAULT_LEDGER: address}
    assert identity.public_key == public_key
    assert identity.public_keys == {DEFAULT_LEDGER: public_key}
    assert identity.default_address_key == DEFAULT_LEDGER



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 open_aea
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test dialogues module for signing protocol."""

# pylint: disable=too-many-statements,too-many-locals,no-member,too-few-public-methods,redefined-builtin
from aea.test_tools.test_protocol import BaseProtocolDialoguesTestCase

from packages.open_aea.protocols.signing.custom_types import RawTransaction, Terms
from packages.open_aea.protocols.signing.dialogues import (
    SigningDialogue,
    SigningDialogues,
)
from packages.open_aea.protocols.signing.message import SigningMessage


class TestDialoguesSigning(BaseProtocolDialoguesTestCase):
    """Test for the 'signing' protocol dialogues."""

    MESSAGE_CLASS = SigningMessage

    DIALOGUE_CLASS = SigningDialogue

    DIALOGUES_CLASS = SigningDialogues

    ROLE_FOR_THE_FIRST_MESSAGE = SigningDialogue.Role.DECISION_MAKER  # CHECK

    def make_message_content(self) -> dict:
        """Make a dict with message contruction content for dialogues.create."""
        return dict(
            performative=SigningMessage.Performative.SIGN_TRANSACTION,
            terms=Terms(
                ledger_id="ledger_id",
                sender_address="address1",
                counterparty_address="address2",
                amount_by_currency_id={"FET": -2},
                quantities_by_good_id={"good_id": 10},
                is_sender_payable_tx_fee=True,
                nonce="transaction nonce",
            ),
            raw_transaction=RawTransaction("ledger_id", {"tx": "transaction"}),
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the implementation of `benchmark` cli command."""
import collections
import sys
from collections import namedtuple
from copy import deepcopy
from typing import Dict, List, Optional

import click
from aea_cli_benchmark.case_acn_communication.command import (
    main as case_acn_communication,
)
from aea_cli_benchmark.case_acn_startup.command import main as case_acn_startup
from aea_cli_benchmark.case_agent_construction_time.command import (
    main as case_agent_construction_time,
)
from aea_cli_benchmark.case_decision_maker.command import main as case_decision_maker
from aea_cli_benchmark.case_dialogues_memory_usage.command import (
    main as case_dialogues_memory_usage,
)
from aea_cli_benchmark.case_mem_usage.command import main as case_mem_usage
from aea_cli_benchmark.case_messages_memory_usage.command import (
    main as case_messages_memory_usage,
)
from aea_cli_benchmark.case_multiagent.command import main as case_mulltiagent
from aea_cli_benchmark.case_multiagent_http_dialogues.command import (
    main as case_multiagent_http_dialogues,
)
from aea_cli_benchmark.case_proactive.command import main as case_proactive
from aea_cli_benchmark.case_reactive.command import main as case_reactive
from aea_cli_benchmark.case_tx_generate.command import main as case_tx_generate

from aea.helpers.yaml_utils import yaml_dump_all, yaml_load_all


@click.group()
@click.pass_context
def benchmark(click_context: click.Context) -> None:  # pylint: disable=unused-argument
    """Run one of performance benchmark."""


benchmark.add_command(case_agent_construction_time)

benchmark.add_command(case_decision_maker)

benchmark.add_command(case_dialogues_memory_usage)

benchmark.add_command(case_mem_usage)

benchmark.add_command(case_messages_memory_usage)

benchmark.add_command(case_multiagent_http_dialogues)

benchmark.add_command(case_mulltiagent)

benchmark.add_command(case_proactive)

benchmark.add_command(case_reactive)
benchmark.add_command(case_tx_generate)
benchmark.add_command(case_acn_startup)
benchmark.add_command(case_acn_communication)

INTERNAL_COMMANDS = ["make-config", "run"]


BenchmarkCase = namedtuple("BenchmarkCase", ["command", "params", "name"])


@click.command(name="run")
@click.argument(
    "file",
    metavar="FILE",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
    required=False,
)
def run(file: Optional[str]):
    """Run benchmarks."""
    if file is not None:
        with open(file, "r") as f:
            cases = _make_cases_from_dicts(yaml_load_all(f))
    else:
        cases = _enlist_cases()
    for case in cases.values():
        click.echo(f"Run case: {case.name}")
        case.command.callback(**case.params)
        click.echo("\n")


def _make_cases_from_dicts(case_definitions: List[Dict]) -> List[BenchmarkCase]:
    base_cases = _enlist_cases()
    case_definitions = deepcopy(case_definitions)
    cases = {}

    for case_def in case_definitions:
        case_name = case_def.pop("name")
        if case_name not in base_cases:
            raise ValueError(f"Unknown case named: {case_name}")
        cases[case_name] = BenchmarkCase(
            name=case_name, params=case_def, command=base_cases[case_name].command
        )
    return cases


def _enlist_cases() -> Dict[str, BenchmarkCase]:
    cases = {}
    for name, command in benchmark.commands.items():
        if name in INTERNAL_COMMANDS:
            continue
        cases[name] = BenchmarkCase(
            command=command,
            name=name,
            params={i.name: i.default for i in command.params},
        )

    return cases


@click.command(name="make-config")
@click.argument(
    "file",
    metavar="FILE",
    type=click.Path(file_okay=True, dir_okay=False),
    required=False,
)
def make_config(file: Optional[str]):
    """Make an example config."""

    def make_section(case: BenchmarkCase) -> Dict:
        result = collections.OrderedDict()
        result["name"] = case.name
        for k, v in sorted(case.params.items(), key=lambda x: x[0]):
            result[k] = v
        return result

    configs = [make_section(case) for case in _enlist_cases().values()]

    if file is not None:
        with open(file, "w") as f:
            yaml_dump_all(configs, stream=f)
    else:
        yaml_dump_all(
            configs,
            stream=sys.stdout,
        )


benchmark.add_command(run)
benchmark.add_command(make_config)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains AEA cli tests for Libp2p tcp client connection."""

# pylint: skip-file

from aea.test_tools.network import LOCALHOST

from packages.valory.connections import p2p_libp2p_mailbox
from packages.valory.connections.p2p_libp2p.tests.base import ports
from packages.valory.connections.p2p_libp2p_mailbox.connection import PUBLIC_ID
from packages.valory.connections.test_libp2p.tests.test_p2p_libp2p_client.test_aea_cli import (
    TestP2PLibp2pClientConnectionAEARunning as Base,
)


DEFAULT_HOST = LOCALHOST.hostname


class TestP2PLibp2pMailboxConnectionAEARunning(Base):
    """Test AEA with p2p_libp2p_client connection is correctly run"""

    conn_path = (
        p2p_libp2p_mailbox_path
    ) = f"vendor.{p2p_libp2p_mailbox.__name__.split('.', 1)[-1]}"
    public_id = str(PUBLIC_ID)
    port = next(ports)
    uri = f"{DEFAULT_HOST}:{port}"
    kwargs = dict(
        delegate_host=DEFAULT_HOST,
        mailbox_port=port,
        delegate=True,
        mailbox=True,
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Scaffold connection and channel."""

import asyncio
from typing import Any, Dict, Optional

from aea.configurations.base import PublicId
from aea.connections.base import Connection, ConnectionStates
from aea.mail.base import Envelope
from aea.protocols.base import Message

from packages.valory.connections.ledger.base import RequestDispatcher
from packages.valory.connections.ledger.contract_dispatcher import (
    ContractApiRequestDispatcher,
)
from packages.valory.connections.ledger.ledger_dispatcher import (
    LedgerApiRequestDispatcher,
)
from packages.valory.protocols.contract_api import ContractApiMessage
from packages.valory.protocols.ledger_api import LedgerApiMessage


PUBLIC_ID = PublicId.from_str("valory/ledger:0.19.0")


class LedgerConnection(Connection):
    """Proxy to the functionality of the SDK or API."""

    connection_id = PUBLIC_ID
    TIMEOUT = 3
    MAX_ATTEMPTS = 120

    def __init__(self, **kwargs: Any):
        """Initialize a connection to interact with a ledger APIs."""
        super().__init__(**kwargs)

        self._ledger_dispatcher: Optional[LedgerApiRequestDispatcher] = None
        self._contract_dispatcher: Optional[ContractApiRequestDispatcher] = None
        self._response_envelopes: Optional[asyncio.Queue] = None

        self.task_to_request: Dict[asyncio.Future, Envelope] = {}
        self.api_configs = self.configuration.config.get(
            "ledger_apis", {}
        )  # type: Dict[str, Dict[str, str]]
        self.request_retry_attempts = self.configuration.config.get(
            "retry_attempts", self.MAX_ATTEMPTS
        )
        self.request_retry_timeout = self.configuration.config.get(
            "retry_timeout", self.TIMEOUT
        )

    @property
    def response_envelopes(self) -> asyncio.Queue:
        """Get the response envelopes. Only intended to be accessed when connected."""
        if self._response_envelopes is None:
            raise ValueError(
                "`asyncio.Queue` for `_response_envelopes` not set. Is the ledger connection active?"
            )
        return self._response_envelopes

    async def connect(self) -> None:
        """Set up the connection."""

        if self.is_connected:  # pragma: nocover
            return

        self.state = ConnectionStates.connecting

        self._ledger_dispatcher = LedgerApiRequestDispatcher(
            self._state,
            loop=self.loop,
            api_configs=self.api_configs,
            logger=self.logger,
            retry_attempts=self.request_retry_attempts,
            retry_timeout=self.request_retry_timeout,
            connection_id=self.connection_id,
        )
        self._contract_dispatcher = ContractApiRequestDispatcher(
            self._state,
            loop=self.loop,
            api_configs=self.api_configs,
            logger=self.logger,
            retry_attempts=self.request_retry_attempts,
            retry_timeout=self.request_retry_timeout,
            connection_id=self.connection_id,
        )

        self._response_envelopes = asyncio.Queue()
        self.state = ConnectionStates.connected

    async def disconnect(self) -> None:
        """Tear down the connection."""
        if self.is_disconnected:  # pragma: nocover
            return

        self.state = ConnectionStates.disconnecting

        for task in self.task_to_request.keys():
            if not task.cancelled():  # pragma: nocover
                task.cancel()
        self._ledger_dispatcher = None
        self._contract_dispatcher = None
        self._response_envelopes = None

        self.state = ConnectionStates.disconnected

    async def send(self, envelope: "Envelope") -> None:
        """
        Send an envelope.

        :param envelope: the envelope to send.
        """
        task = self._schedule_request(envelope)
        task.add_done_callback(self._handle_done_task)
        self.task_to_request[task] = envelope

    def _schedule_request(self, envelope: Envelope) -> asyncio.Task:
        """
        Schedule a ledger API request.

        :param envelope: the message.
        :return: task
        """
        dispatcher: RequestDispatcher
        if (
            envelope.protocol_specification_id
            == LedgerApiMessage.protocol_specification_id
        ):
            if self._ledger_dispatcher is None:  # pragma: nocover
                raise ValueError("No ledger dispatcher set.")
            dispatcher = self._ledger_dispatcher
        elif (
            envelope.protocol_specification_id
            == ContractApiMessage.protocol_specification_id
        ):
            if self._contract_dispatcher is None:  # pragma: nocover
                raise ValueError("No contract dispatcher set.")
            dispatcher = self._contract_dispatcher
        else:
            raise ValueError("Protocol not supported")

        task = dispatcher.dispatch(envelope)
        return task

    async def receive(self, *args: Any, **kwargs: Any) -> Optional["Envelope"]:
        """
        Receive an envelope. Blocking.

        :param args: the arguments
        :param kwargs: the keyword arguments
        :return: the envelope received, or None.
        """
        return await self.response_envelopes.get()

    def _handle_done_task(self, task: asyncio.Future) -> None:
        """
        Process a done receiving task.

        :param task: the done task.
        """
        request = self.task_to_request.pop(task)
        response_message: Optional[Message] = task.result()

        response_envelope = None
        if response_message is not None:
            response_envelope = Envelope(
                to=request.sender,
                sender=request.to,
                message=response_message,
                context=request.context,
            )

        # not handling `asyncio.QueueFull` exception, because the maxsize we defined for the Queue is infinite
        self.response_envelopes.put_nowait(response_envelope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains AEA cli tests for P2PLibp2p connection."""

# pylint: skip-file

import json

from aea.configurations.constants import DEFAULT_LEDGER
from aea.test_tools.network import LOCALHOST

from packages.valory.connections.p2p_libp2p.connection import (
    PUBLIC_ID as P2P_CONNECTION_PUBLIC_ID,
)
from packages.valory.connections.p2p_libp2p.tests.base import (
    libp2p_log_on_failure_all,
    ports,
)
from packages.valory.connections.p2p_libp2p.tests.test_aea_cli import (
    BaseP2PLibp2pConnectionAEATest,
    p2p_libp2p_path,
)
from packages.valory.connections.test_libp2p.tests.base import LIBP2P_LEDGER


@libp2p_log_on_failure_all
class TestP2PLibp2pConnectionAEARunningDefaultConfigNode(
    BaseP2PLibp2pConnectionAEATest
):
    """Test AEA with p2p_libp2p connection is correctly run"""

    def test_agent(self) -> None:
        """Test with aea."""
        agent_ledger_id, node_ledger_id = DEFAULT_LEDGER, LIBP2P_LEDGER
        # set config
        self.set_config("agent.default_ledger", agent_ledger_id)
        self.set_config(
            "agent.required_ledgers",
            json.dumps([agent_ledger_id, node_ledger_id]),
            "list",
        )
        self.set_config("agent.default_connection", str(P2P_CONNECTION_PUBLIC_ID))
        # agent keys
        self.generate_private_key(agent_ledger_id)
        self.add_private_key(agent_ledger_id, f"{agent_ledger_id}_private_key.txt")
        # libp2p node keys
        self.generate_private_key(node_ledger_id, private_key_file=self.conn_key_file)
        self.add_private_key(
            node_ledger_id, private_key_filepath=self.conn_key_file, connection=True
        )
        # add connection and build
        self.add_item("connection", str(P2P_CONNECTION_PUBLIC_ID))
        self.run_cli_command("build", cwd=self._get_cwd())
        self.run_cli_command("issue-certificates", cwd=self._get_cwd())

        self.set_logging()
        self.run_aea_cli_test()


@libp2p_log_on_failure_all
class TestP2PLibp2pConnectionAEARunningFullNode(BaseP2PLibp2pConnectionAEATest):
    """Test AEA with p2p_libp2p connection is correctly run"""

    def test_agent(self) -> None:
        """Test with aea."""
        agent_ledger_id, node_ledger_id = DEFAULT_LEDGER, LIBP2P_LEDGER
        # set config
        self.set_config("agent.default_ledger", agent_ledger_id)
        self.set_config(
            "agent.required_ledgers",
            json.dumps([agent_ledger_id, node_ledger_id]),
            "list",
        )
        # agent keys
        self.generate_private_key(agent_ledger_id)
        self.add_private_key(agent_ledger_id, f"{agent_ledger_id}_private_key.txt")
        # libp2p node keys
        self.generate_private_key(node_ledger_id, private_key_file=self.conn_key_file)
        self.add_private_key(
            node_ledger_id, private_key_filepath=self.conn_key_file, connection=True
        )
        # add connection and build
        self.add_item("connection", str(P2P_CONNECTION_PUBLIC_ID))
        self.run_cli_command("build", cwd=self._get_cwd())

        # setup a full node: with public uri, relay service, and delegate service
        config_path = f"{p2p_libp2p_path}.config"
        hostname = LOCALHOST.hostname
        self.set_config(f"{config_path}.local_uri", f"{hostname}:{next(ports)}")
        self.set_config(f"{config_path}.public_uri", f"{hostname}:{next(ports)}")
        self.set_config(f"{config_path}.delegate_uri", f"{hostname}:{next(ports)}")

        self.run_cli_command("issue-certificates", cwd=self._get_cwd())

        self.set_logging()
        self.run_aea_cli_test()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Module wrapping the helpers of public and private key cryptography."""
import logging
import os
from pathlib import Path
from typing import Dict, Optional, Union

from aea.configurations.base import AgentConfig
from aea.configurations.constants import PRIVATE_KEY_PATH_SCHEMA
from aea.crypto.registries import crypto_registry, make_crypto, make_faucet_api
from aea.crypto.wallet import Wallet
from aea.helpers.base import ensure_dir
from aea.helpers.env_vars import is_env_variable


_default_logger = logging.getLogger(__name__)

_ = PRIVATE_KEY_PATH_SCHEMA  # some modules expect this here


def try_validate_private_key_path(
    ledger_id: str,
    private_key_path: str,
    password: Optional[str] = None,
) -> None:
    """
    Try validate a private key path.

    :param ledger_id: one of 'fetchai', 'ethereum'
    :param private_key_path: the path to the private key.
    :param password: the password to encrypt/decrypt the private key.
    """
    try:
        # to validate the file, we just try to create a crypto object
        # with private_key_path as parameter
        make_crypto(ledger_id, private_key_path=private_key_path, password=password)
    except Exception as e:  # pylint: disable=broad-except  # thats ok, reraise
        error_msg = (
            "This is not a valid private key file: '{}'\n Exception: '{}'".format(
                private_key_path, e
            )
        )
        _default_logger.error(error_msg)
        raise


def create_private_key(
    ledger_id: str,
    private_key_file: str,
    password: Optional[str] = None,
    extra_entropy: Union[str, bytes, int] = "",
) -> None:
    """
    Create a private key for the specified ledger identifier.

    :param ledger_id: the ledger identifier.
    :param private_key_file: the private key file.
    :param password: the password to encrypt/decrypt the private key.
    :param extra_entropy: add extra randomness to whatever randomness your OS can provide
    """
    crypto = make_crypto(ledger_id, extra_entropy=extra_entropy)
    crypto.dump(private_key_file, password)


def try_generate_testnet_wealth(
    identifier: str, address: str, url: Optional[str] = None, _sync: bool = True
) -> None:
    """
    Try generate wealth on a testnet.

    :param identifier: the identifier of the ledger
    :param address: the address to check for
    :param url: the url
    :param _sync: whether to wait to sync or not; currently unused
    """
    faucet_api = make_faucet_api(identifier)
    if faucet_api is not None:
        faucet_api.get_wealth(address, url)


def private_key_verify(
    aea_conf: AgentConfig,
    aea_project_path: Path,
    password: Optional[str] = None,
) -> None:
    """
    Check key.

    :param aea_conf: AgentConfig
    :param aea_project_path: Path, where project placed.
    :param password: the password to encrypt/decrypt the private key.
    """
    for identifier, _ in aea_conf.private_key_paths.read_all():
        if identifier not in crypto_registry.supported_ids:  # pragma: nocover
            raise ValueError(
                "Unsupported identifier `{}` in private key paths. Supported identifiers: {}.".format(
                    identifier, sorted(crypto_registry.supported_ids)
                )
            )

    for identifier in crypto_registry.supported_ids:
        config_private_key_path = aea_conf.private_key_paths.read(identifier)

        if is_env_variable(config_private_key_path):
            # config_private_key_path is env variable to be used, skip it. check will be performed after substitution
            continue

        if config_private_key_path is None:
            continue
        try:
            try_validate_private_key_path(
                identifier,
                str(aea_project_path / config_private_key_path),
                password=password,
            )
        except FileNotFoundError:  # pragma: no cover
            raise ValueError(
                "File {} for private key {} not found.".format(
                    repr(config_private_key_path),
                    identifier,
                )
            )


def make_certificate(
    ledger_id: str,
    crypto_private_key_path: str,
    message: bytes,
    output_path: str,
    password: Optional[str] = None,
) -> str:
    """
    Create certificate.

    :param ledger_id: the ledger id
    :param crypto_private_key_path: the path to the private key.
    :param message: the message to be signed.
    :param output_path: the location where to save the certificate.
    :param password: the password to encrypt/decrypt the private keys.
    :return: the signature/certificate
    """
    crypto = crypto_registry.make(
        ledger_id, private_key_path=crypto_private_key_path, password=password
    )
    signature = crypto.sign_message(message).encode("ascii").hex()
    ensure_dir(os.path.dirname(output_path))
    Path(output_path).write_bytes(signature.encode("ascii"))
    return signature


def get_wallet_from_agent_config(
    agent_config: AgentConfig, password: Optional[str] = None
) -> Wallet:
    """
    Get wallet from agent_cofig provided.

    :param agent_config: the agent configuration object
    :param password: the password to encrypt/decrypt the private keys.
    :return: wallet
    """
    private_key_paths: Dict[str, Optional[str]] = {
        config_pair[0]: config_pair[1]
        for config_pair in agent_config.private_key_paths.read_all()
    }
    connections_private_key_paths: Dict[str, Optional[str]] = {
        config_pair[0]: config_pair[1]
        for config_pair in agent_config.connection_private_key_paths.read_all()
    }
    wallet = Wallet(private_key_paths, connections_private_key_paths, password=password)
    return wallet


class DecryptError(ValueError):
    """Error on bytes decryption with password."""

    msg = "Decrypt error! Bad password?"

    def __init__(self, msg: Optional[str] = None) -> None:
        """Init exception."""
        super().__init__(msg or self.msg)


class KeyIsIncorrect(ValueError):
    """Error decoding hex string to bytes for private key."""


def hex_to_bytes_for_key(data: str) -> bytes:
    """Convert hex string to bytes with error handling."""
    try:
        return bytes.fromhex(data)
    except ValueError as e:
        raise KeyIsIncorrect(str(e)) from e
